// $Id: ConfigurationDialog.cpp,v 1.18 2013/03/12 08:47:46 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "ConfigurationDialog.h"

#include "Application.h"
#include "Debug.h"
#include "GridLayout.h"
#include "Icons.h"
#include "IconEngine.h"
#include "ImageRecordConfiguration.h"
#include "ImageRecordList.h"
#include "OptionCheckBox.h"
#include "OptionColorDisplay.h"
#include "OptionLineEditor.h"
#include "OptionSpinBox.h"
#include "Singleton.h"

#include <QGroupBox>
#include <QLabel>
#include <QDoubleValidator>

//_________________________________________________________
ConfigurationDialog::ConfigurationDialog( QWidget* parent ):
    BaseConfigurationDialog( parent )
{
    Debug::Throw( "ConfigurationDialog::ConfigurationDialog.\n" );
    setWindowTitle( tr( "Configuration - Qedit" ) );

    // base configuration
    baseConfiguration();


    // display
    QWidget* page = &addPage( IconEngine::get( ICONS::PREFERENCE_APPEARANCE ), tr( "Appearance" ), tr( "Markers appearance" ) );

    {
        QGroupBox* box = new QGroupBox( tr( "Colors" ), page );
        page->layout()->addWidget( box );

        QVBoxLayout* vLayout( new QVBoxLayout() );
        box->setLayout( vLayout );

        GridLayout* gridLayout = new GridLayout();
        gridLayout->setMaxCount(2);
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
        vLayout->addLayout( gridLayout );

        OptionColorDisplay* colorDisplay;
        QLabel* label;
        gridLayout->addWidget( label = new QLabel( tr( "Background color:" ), box ) );
        gridLayout->addWidget( colorDisplay = new OptionColorDisplay( box, "SCROLL_AREA_COLOR" ) );
        addOptionWidget( colorDisplay );
        label->setBuddy( colorDisplay );

        gridLayout->addWidget( label = new QLabel( tr( "Reference points color:" ), box ) );
        gridLayout->addWidget( colorDisplay = new OptionColorDisplay( box, "REF_POINT_COLOR" ) );
        addOptionWidget( colorDisplay );
        label->setBuddy( colorDisplay );

        gridLayout->addWidget( label = new QLabel( tr( "Markers color:" ), box ) );
        gridLayout->addWidget( colorDisplay = new OptionColorDisplay( box, "POINT_COLOR" ) );
        addOptionWidget( colorDisplay );
        label->setBuddy( colorDisplay );

        gridLayout->addWidget( label = new QLabel( tr( "Inactive markers color:" ), box ) );
        gridLayout->addWidget( colorDisplay = new OptionColorDisplay( box, "INACTIVE_POINT_COLOR" ) );
        addOptionWidget( colorDisplay );
        label->setBuddy( colorDisplay );

        gridLayout->addWidget( label = new QLabel( tr( "Selected markers color:" ), box ) );
        gridLayout->addWidget( colorDisplay = new OptionColorDisplay( box, "SELECTED_POINT_COLOR" ) );
        addOptionWidget( colorDisplay );
        label->setBuddy( colorDisplay );

        gridLayout->addWidget( label = new QLabel( tr( "Hovered markers color:" ), box ) );
        gridLayout->addWidget( colorDisplay = new OptionColorDisplay( box, "HOVER_POINT_COLOR" ) );
        addOptionWidget( colorDisplay );
        label->setBuddy( colorDisplay );
    }

    {
        QGroupBox* box = new QGroupBox( tr( "Options" ), page );
        page->layout()->addWidget( box );

        QVBoxLayout* vLayout( new QVBoxLayout() );
        box->setLayout( vLayout );

        GridLayout* gridLayout = new GridLayout();
        gridLayout->setMaxCount(2);
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
        vLayout->addLayout( gridLayout );

        QLabel* label;
        OptionSpinBox* spinbox;
        gridLayout->addWidget( label = new QLabel( tr( "Marker size:" ), box ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( box, "POINT_SIZE" ) );
        spinbox->setSuffix( tr( "px" ) );
        spinbox->setMinimum( 5 );
        spinbox->setMaximum( 50 );
        addOptionWidget( spinbox );
        label->setBuddy( spinbox );

        gridLayout->addWidget( label = new QLabel( tr( "Marker line width:" ), box ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( box, "PEN_SIZE" ) );
        spinbox->setSuffix( tr( "px" ) );
        spinbox->setMinimum( 1 );
        spinbox->setMaximum( 5 );
        addOptionWidget( spinbox );
        label->setBuddy( spinbox );


        OptionCheckBox* checkBox;
        vLayout->addWidget( checkBox = new OptionCheckBox( tr( "Show reference points" ), box, "SHOW_REF_POINTS" ) );
        addOptionWidget( checkBox );

        vLayout->addWidget( checkBox = new OptionCheckBox( tr( "Show inactive markers" ), box, "SHOW_INACTIVE_POINTS" ) );
        addOptionWidget( checkBox );

        vLayout->addWidget( checkBox = new OptionCheckBox( tr( "Show scan lines" ), box, "SHOW_SCAN_LINES" ) );
        addOptionWidget( checkBox );

    }

    {
        QGroupBox* box = new QGroupBox( tr( "Default Errors (fractional)" ), page );
        page->layout()->addWidget( box );

        QVBoxLayout* vLayout( new QVBoxLayout() );
        box->setLayout( vLayout );

        GridLayout* gridLayout = new GridLayout();
        gridLayout->setMaxCount(2);
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
        vLayout->addLayout( gridLayout );

        QLabel* label;
        OptionLineEditor* editor;
        gridLayout->addWidget( label = new QLabel( tr( "Horizontal statistical error:" ), box ) );
        gridLayout->addWidget( editor = new OptionLineEditor( box, "DEFAULT_ERROR_X" ) );
        editor->setValidator( new QDoubleValidator() );
        addOptionWidget( editor );
        label->setBuddy( editor );

        gridLayout->addWidget( label = new QLabel( tr( "Vertical statistical error:" ), box ) );
        gridLayout->addWidget( editor = new OptionLineEditor( box, "DEFAULT_ERROR_Y" ) );
        editor->setValidator( new QDoubleValidator() );
        addOptionWidget( editor );
        label->setBuddy( editor );

        gridLayout->addWidget( new QLabel( tr( "Vertical systematic error:" ), box ) );
        gridLayout->addWidget( editor = new OptionLineEditor( box, "DEFAULT_ERROR_SYST_Y" ) );
        editor->setValidator( new QDoubleValidator() );
        addOptionWidget( editor );
    }

    // notifications
    page = &addPage( IconEngine::get( ICONS::PREFERENCE_NOTIFICATIONS ), tr( "Notifications" ),tr(  "Notifications" ) );
    {

        QVBoxLayout* vLayout( new QVBoxLayout() );
        vLayout->setSpacing(5);
        page->layout()->addItem( vLayout );

        OptionCheckBox* checkBox;
        vLayout->addWidget( checkBox = new OptionCheckBox( tr( "Show notifications" ), page, "SHOW_NOTIFICATIONS" ) );
        addOptionWidget( checkBox );

        vLayout->addWidget( checkBox = new OptionCheckBox( tr( "Animate notifications" ), page, "ANIMATE_NOTIFICATIONS" ) );
        addOptionWidget( checkBox );

        vLayout->addWidget( checkBox = new OptionCheckBox( tr( "Auto-hide notifications" ), page, "AUTO_HIDE_NOTIFICATIONS" ) );
        addOptionWidget( checkBox );

        GridLayout* gridLayout = new GridLayout();
        gridLayout->setMaxCount(2);
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
        vLayout->addLayout( gridLayout );

        OptionSpinBox* spinbox;
        gridLayout->addWidget( new QLabel( tr( "Animation duration:" ), page ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( page, "ANIMATION_DURATION" ) );
        spinbox->setSuffix( tr( "ms" ) );
        spinbox->setMinimum( 0 );
        spinbox->setMaximum( 2000 );
        addOptionWidget( spinbox );

        gridLayout->addWidget( new QLabel( tr( "Auto-hide delay:" ), page ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( page, "AUTO_HIDE_DELAY" ) );
        spinbox->setSuffix( tr( "ms" ) );
        spinbox->setMinimum( 0 );
        spinbox->setMaximum( 10000 );
        addOptionWidget( spinbox );

        gridLayout->setColumnStretch( 1, 1 );

    }

    // notifications
    page = &addPage( IconEngine::get( ICONS::PREFERENCE_RECENT_FILES ), tr( "Recent Files" ), tr( "Recent files" ) );
    {
        ImageRecordConfiguration* imageRecordConfiguration = new ImageRecordConfiguration( page, Singleton::get().application<Application>()->recentImages() );
        page->layout()->addWidget( imageRecordConfiguration );
        addOptionWidget( imageRecordConfiguration );
    }
    // load initial configuration
    read();

}
