// $Id: DisplayWidget.cpp,v 1.48 2013/03/11 16:52:44 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "DisplayWidget.h"

#include "BaseContextMenu.h"
#include "Color.h"
#include "IconEngine.h"
#include "Icons.h"
#include "InformationDialog.h"
#include "Singleton.h"
#include "XmlOptions.h"

#include <QAbstractScrollArea>
#include <QAction>
#include <QApplication>
#include <QMouseEvent>
#include <QPainter>
#include <QScrollBar>
#include <QShortcut>
#include <QString>
#include <QTextStream>

#include <cmath>

//________________________________________________________________________
DisplayWidget::DisplayWidget( QWidget* parent ):
    QWidget( parent ),
    Counter( "DisplayWidget" ),
    refPointsValid_( false ),
    dataSetIndex_( 0 ),
    pointSize_( 20 ),
    penWidth_( 1 ),
    zoomMode_( ZoomWidget::Fixed ),
    zoomFactor_( 1.0 ),
    realZoomFactor_( 1.0 ),
    hoverMode_( RefPoint::NoHover ),
    dragMode_( RefPoint::NoDrag )
{
    setAttribute( Qt::WA_OpaquePaintEvent, true );
    setAttribute( Qt::WA_Hover, true );
    setMouseTracking( true );
    setCursor( Qt::CrossCursor );

    // remove selected point shortcut
    new QShortcut( QKeySequence::Delete, this, SLOT( _removeSelection( void ) ) );

    // actions
    _installActions();

    // configuration
    connect( Singleton::get().application(), SIGNAL( configurationChanged() ), SLOT( _updateConfiguration() ) );
    _updateConfiguration();

}

//________________________________________________________________________
void DisplayWidget::setImage( const File& filename )
{

    file_ = filename;
    pixmap_ = QPixmap( filename );
    _adjustZoomFactor();
    _updateScaledPixmap();
    update();

    if( pixmap_.isNull() )
    {

        InformationDialog( this, QString( tr( "Unable to load image from file '%1'" ) ).arg( filename ) ).exec();
        file_ = File();

    }

}

//________________________________________________________________________
void DisplayWidget::zoomModified( ZoomWidget::Mode mode, double factor )
{
    Debug::Throw() << "DisplayWidget::zoomModified - mode: " << mode << " scale: " << factor << endl;

    // make sure things are changed
    if( mode == zoomMode_ && factor == zoomFactor_ ) return;
    zoomMode_ = mode;
    if( zoomMode_ == ZoomWidget::Fixed )
    {

        zoomFactor_ = factor;

    } else if( !pixmap_.isNull() ) {

        _adjustZoomFactor();

    }

    _updateScaledPixmap();
    _updateHoverPoint();
    updatePosition();

}

//________________________________________________________________________
void DisplayWidget::updatePosition()
{ setPosition( mapFromGlobal( QCursor::pos() ) ); }

//________________________________________________________________________
bool DisplayWidget::setPosition( QPoint position )
{
    if( position != position_ )
    {
        position_ = position;
        if( _updateHoverPoint() || showScanLinesAction().isChecked() )
        { update(); }

        return true;
    } else return false;
}

//________________________________________________________________________
bool DisplayWidget::eventFilter( QObject*, QEvent* event )
{
    if( zoomMode_ == ZoomWidget::Fixed ) return false;

    switch( event->type() )
    {
        case QEvent::Resize:
        resizeTimer_.start( 200, this );
        break;

        default: break;
    }

    return false;
}

//________________________________________________________________________
bool DisplayWidget::positionIsValid( void )
{
    if( !rect().contains( position_ ) ) return false;
    else if( parentWidget() ) return parentWidget()->rect().contains( mapToParent( position_ ) );
    else return false;
}

//________________________________________________________________________
bool DisplayWidget::event( QEvent* event )
{

    switch( event->type() )
    {

        case QEvent::HoverEnter:
        {
            QPoint position( static_cast<QHoverEvent*>(event)->pos() );
            setPosition( position );
            emit widgetEntered();
            return true;
        }

        case QEvent::HoverLeave:
        {
            updatePosition();
            emit widgetLeft();
            return true;
        }

        default: return QWidget::event( event );
    }
}

//________________________________________________________________________
void DisplayWidget::paintEvent( QPaintEvent* event )
{

    Debug::Throw( "DisplayWidget::paintEvent.\n" );

    QPainter painter( this );
    painter.setClipRegion( event->region() );
    painter.setPen( Qt::NoPen );
    if( scaledPixmap_.isNull() )
    {

        painter.setBrush( palette().color( QPalette::Window ) );
        painter.drawRect( rect() );
        return;
    }

    painter.drawPixmap( QPoint(0,0), scaledPixmap_ );

    if( showScanLinesAction().isChecked() && positionIsValid() )
    {

        // draw scanlines
        painter.setPen( QPen( Qt::black, 1, Qt::DotLine ) );
        painter.drawLine( QPoint( rect().left(), position_.y() ), QPoint( rect().right()+1, position_.y() ) );
        painter.drawLine( QPoint( position_.x(), rect().top() ), QPoint( position_.x(), rect().bottom() ) );

    }

    // draw symbols
    painter.setRenderHints( QPainter::Antialiasing );
    painter.setBrush( Qt::NoBrush );

    // reference points
    if( showRefPointsAction().isChecked() || !refPointsValid_ )
    {
        painter.setPen( QPen( refPointColor_, penWidth_ ) );
        for( RefPoint::Set::const_iterator iter = refPoints_.begin(); iter != refPoints_.end(); ++iter )
        {
            if( selectedPoints_.find( *iter ) != selectedPoints_.end() ) continue;
            _drawSymbol( painter, *iter );
        }
    }

    // points
    if( showInactivePointsAction().isChecked() )
    {
        painter.setPen( QPen( inactivePointColor_, penWidth_ ) );
        for( int index = 0; index < (int) dataSets_.size(); index++ )
        {
            // ignore current dataSet
            if( index == dataSetIndex_ ) continue;

            // draw all others as inactive
            const RefPoint::List& points( dataSets_[index] );
            for( RefPoint::List::const_iterator iter = points.begin(); iter != points.end(); ++iter )
            {
                if( hoverMode_ != RefPoint::NoHover && hoverPoint_ == *iter ) continue;
                _drawSymbol( painter, *iter );
            }

        }
    }

    // current dataSet
    painter.setPen( QPen( pointColor_, penWidth_ ) );
    for( RefPoint::Set::const_iterator iter = points_.begin(); iter != points_.end(); ++iter )
    {
        if( selectedPoints_.find( *iter ) != selectedPoints_.end() ) continue;
        if( hoverMode_ != RefPoint::NoHover && hoverPoint_ == *iter ) continue;
        _drawSymbol( painter, *iter );
    }

    // selected points
    painter.setPen( QPen( selectedPointColor_, penWidth_ ) );
    for( RefPoint::Set::const_iterator iter = selectedPoints_.begin(); iter != selectedPoints_.end(); ++iter )
    {

        if( refPoints_.find( *iter ) != refPoints_.end() )
        {
            const QPointF position( iter->global()*realZoomFactor_ );

            // draw scanlines
            painter.save();
            painter.setRenderHint( QPainter::Antialiasing, false );
            painter.setPen( QPen( selectedPointColor_, 1, Qt::DotLine ) );
            painter.drawLine( QPoint( rect().left(), position.y() ), QPoint( rect().right()+1, position.y() ) );
            painter.drawLine( QPoint( position.x(), rect().top() ), QPoint( position.x(), rect().bottom() ) );
            painter.restore();
        }

        if( hoverMode_ != RefPoint::NoHover && hoverPoint_ == *iter ) continue;
        _drawSymbol( painter, *iter );
    }

    // hover
    if( hoverMode_ != RefPoint::NoHover )
    {
        if( selectedPoints_.find( hoverPoint_ ) != selectedPoints_.end() ) painter.setPen( QPen( selectedPointColor_, penWidth_ ) );
        else if( points_.find( hoverPoint_ ) != points_.end() ) painter.setPen( QPen( pointColor_, penWidth_ ) );
        else painter.setPen( QPen( inactivePointColor_, penWidth_ ) );
        _drawSymbol( painter, hoverPoint_, hoverMode_ );
    }

    // drag
    if( dragMode_ != RefPoint::NoDrag )
    {
        painter.setPen( QPen( hoverPointColor_, penWidth_ ) );
        _drawSymbol( painter, dragPoint_ );
    }

}

//________________________________________________________________________
void DisplayWidget::timerEvent( QTimerEvent* event )
{

    Debug::Throw( "DisplayWidget::timerEvent.\n" );
    if( event->timerId() != resizeTimer_.timerId() || zoomMode_ == ZoomWidget::Fixed )
    { return QWidget::timerEvent( event ); }

    resizeTimer_.stop();
    _adjustZoomFactor();
    _updateScaledPixmap();
    update();

}

//________________________________________________________________________
void DisplayWidget::mouseMoveEvent( QMouseEvent* event )
{

    QPoint position( event->pos() );

    // update position, and hover (for scan lines)
    setPosition( position );

    if( dragMode_ != RefPoint::NoDrag && ( event->buttons() & Qt::LeftButton ) )
    { emit pointDragged( QPointF( event->pos() )*(1.0/realZoomFactor_), dragMode_ ); }

    // emit signal
    emit pointerMoved( position, hoverMode_ );

}

//________________________________________________________________________
void DisplayWidget::mousePressEvent( QMouseEvent* event )
{
    Debug::Throw( "DisplayWidget::mousePressEvent.\n" );

    if( event->button() == Qt::LeftButton && event->modifiers() == Qt::NoModifier)
    {

        clickPosition_ = event->pos();

        if( hoverMode_ == RefPoint::NoHover ) return;

        // emit signal
        emit pointSelected();

        // can only select one ref point at a time
        if( hoverMode_ == RefPoint::HoverRefPoint )
        { selectedPoints_.clear(); }

        // toggle drag in progress
        /* FIXME: this call is dangerous. Better use typedef and a unique enum */
        dragMode_ = (RefPoint::DragMode) hoverMode_;

        // store
        dragPoint_ = hoverPoint_;

        /*
        remove iterator from points
        it will be re-inserted when drag is finished
        */
        selectedPoints_.remove( dragPoint_ );
        points_.remove( dragPoint_ );

        update();

    } else if( event->button() == Qt::RightButton && hoverMode_ != RefPoint::NoHover ) {

        emit pointSelected();

    }

}

//________________________________________________________________________
void DisplayWidget::mouseReleaseEvent( QMouseEvent* event )
{

    Debug::Throw( "DisplayWidget::mouseReleaseEvent.\n" );

    if( event->button() == Qt::LeftButton )
    {

        if( dragMode_ == RefPoint::NoDrag && (clickPosition_ - event->pos()).manhattanLength() < qApp->startDragDistance() )
        {

            emit pointAdded( QPointF( event->pos() )*(1.0/realZoomFactor_) );

        } else {

            // copy drag mode on hover mode
            hoverMode_ = (RefPoint::HoverMode) dragMode_;
            hoverPoint_ = dragPoint_;

            // reset drag mode
            dragMode_ = RefPoint::NoDrag;

            // re-insert drag point in lists
            if( hoverMode_ != RefPoint::HoverRefPoint )
            { points_.insert( dragPoint_ ); }

            // re-insert in list of selected points
            selectedPoints_.insert( dragPoint_ );

        }

    }

    return;

}

//________________________________________________________________________
void DisplayWidget::contextMenuEvent( QContextMenuEvent* event )
{

    // update actions text and enable state
    removePointAction_->setEnabled( hoverMode_ == RefPoint::HoverPoint || hoverMode_ == RefPoint::HoverRefPoint );
    if( hoverMode_ == RefPoint::HoverRefPoint ) removePointAction_->setText( tr( "Remove This Reference Point" ) );
    else removePointAction_->setText( tr( "Remove This Marker" ) );

    copyPointAction_->setEnabled( hoverMode_ == RefPoint::HoverPoint );
    editErrorsAction_->setEnabled( hoverMode_ == RefPoint::HoverPoint );

    defaultErrorsAction_->setEnabled(
        hoverMode_ == RefPoint::HoverPoint &&
        ((!hoverPoint_.hasErrorX() && XmlOptions::get().get<double>( "DEFAULT_ERROR_X" ) > 0 ) ||
        (!hoverPoint_.hasErrorY() && XmlOptions::get().get<double>( "DEFAULT_ERROR_Y" ) > 0 ) )
        );

    defaultErrorsSystAction_->setEnabled(
        hoverMode_ == RefPoint::HoverPoint &&
        ((!hoverPoint_.hasErrorSystX() && XmlOptions::get().get<double>( "DEFAULT_ERROR_SYST_X" ) > 0 ) ||
        (!hoverPoint_.hasErrorSystY() && XmlOptions::get().get<double>( "DEFAULT_ERROR_SYST_Y" ) > 0 ) )
        );

    addPointAction_->setText( refPointsValid_ ? tr( "Add Marker" ) : tr( "Add Reference Point" ) );

    BaseContextMenu menu( this );
    menu.setHideDisabledActions( true );

    // update context menu
    menu.addAction( addPointAction_ );
    menu.addSeparator();

    menu.addAction( copyPointAction_ );
    menu.addAction( pastePointAction_ );
    menu.addAction( removePointAction_ );

    menu.addSeparator();
    menu.addAction( editErrorsAction_ );
    menu.addAction( defaultErrorsAction_ );
    menu.addAction( defaultErrorsSystAction_ );

    // configuration
    menu.addSeparator();
    menu.addAction( showRefPointsAction_ );
    menu.addAction( showInactivePointsAction_ );
    menu.addAction( showScanLinesAction_ );

    menu.exec( event->globalPos() );
    updatePosition();
}

//________________________________________________________________________
void DisplayWidget::_installActions( void )
{

    addAction( showRefPointsAction_ = new QAction( tr( "Show Reference Points" ), this ) );
    showRefPointsAction_->setCheckable( true );
    connect( showRefPointsAction_, SIGNAL( toggled( bool ) ), SLOT( _toggleShowRefPoints( bool ) ) );

    addAction( showInactivePointsAction_ = new QAction( tr( "Show Inactive Markers" ), this ) );
    showInactivePointsAction_->setCheckable( true );
    connect( showInactivePointsAction_, SIGNAL( toggled( bool ) ), SLOT( _toggleShowInactivePoints( bool ) ) );

    addAction( showScanLinesAction_ = new QAction( tr( "Show Scan Lines" ), this ) );
    showScanLinesAction_->setCheckable( true );
    connect( showScanLinesAction_, SIGNAL( toggled( bool ) ), SLOT( _toggleShowScanLines( bool ) ) );

    addAction( addPointAction_ = new QAction( IconEngine::get( ICONS::ADD ), tr( "Add Marker" ), this ) );
    connect( addPointAction_, SIGNAL( triggered( void ) ), SLOT( _addPoint( void ) ) );

    addAction( removePointAction_ = new QAction( IconEngine::get( ICONS::REMOVE ), "Remove This Marker", this ) );
    removePointAction_->setShortcut( QKeySequence::Cut );

    addAction( editErrorsAction_ = new QAction( IconEngine::get( ICONS::EDIT ), tr( "Edit This Marker Errors" ), this ) );
    addAction( defaultErrorsAction_ = new QAction( IconEngine::get( ICONS::EDIT ), tr( "Add Default Statistical Errors to This Marker" ), this ) );
    addAction( defaultErrorsSystAction_ = new QAction( IconEngine::get( ICONS::EDIT ), tr( "Add Default Systematic Errors to This Marker" ), this ) );

    addAction( copyPointAction_ = new QAction( IconEngine::get( ICONS::COPY ), tr( "Copy This Marker" ), this ) );
    copyPointAction_->setShortcut( QKeySequence::Copy );

    addAction( pastePointAction_ = new QAction( IconEngine::get( ICONS::PASTE ), tr( "Paste One Marker" ), this ) );
    pastePointAction_->setShortcut( QKeySequence::Paste );
    pastePointAction_->setEnabled( false );
    connect( pastePointAction_, SIGNAL( triggered( void ) ), SLOT( _pastePoint( void ) ) );

}

//________________________________________________________________________
void DisplayWidget::_updateScaledPixmap( void )
{
    if( pixmap_.isNull() ) scaledPixmap_ = QPixmap();
    else {

        if( zoomFactor_ == 1 ) scaledPixmap_ = pixmap_;
        else scaledPixmap_ = pixmap_.scaled( pixmap_.size()*zoomFactor_, Qt::KeepAspectRatio, Qt::SmoothTransformation );
        resize( scaledPixmap_.size() );

        realZoomFactor_ = double( scaledPixmap_.width() )/double( pixmap_.width() );
        emit realZoomFactorChanged( realZoomFactor_ );

    }

    return;
}

//________________________________________________________________________
bool DisplayWidget::_updateHoverPoint( void )
{

    // update hover point
    if( hoverMode_ != RefPoint::NoHover && points_.find( hoverPoint_ ) == points_.end() )
    { hoverMode_ = RefPoint::NoHover; }

    // find nearest point
    RefPoint::HoverMode hoverMode( RefPoint::NoHover );
    const RefPoint point( _findPoint( position_, hoverMode ) );

    // look for changes
    const bool changed( hoverMode != hoverMode_ || (hoverMode != RefPoint::NoHover && !( point == hoverPoint_ ) ) );

    // update
    hoverMode_ = hoverMode;
    if( hoverMode_ != RefPoint::NoHover ) hoverPoint_ = point;

    return changed;

}

//________________________________________________________________________
void DisplayWidget::_drawSymbol( QPainter& painter, const RefPoint& point, const RefPoint::HoverMode& hoverMode ) const
{

    QPen oldPen( painter.pen() );
    QPen hoverPen( hoverPointColor_, oldPen.width() );

    // draw ellipse
    const QPointF scaledPosition( point.global()*realZoomFactor_ );
    const double pointSize( pointSize_*realZoomFactor_ );
    const QRectF rect(  scaledPosition - QPointF( pointSize, pointSize ), QSizeF( pointSize*2, pointSize*2 ) );

    if( hoverMode == RefPoint::HoverPoint || hoverMode == RefPoint::HoverRefPoint ) painter.setPen( hoverPen );
    painter.drawEllipse( rect );

    if( !point.hasErrors() )
    {
        const double radius = pointSize/std::sqrt(2.0);
        painter.drawLine( scaledPosition - QPointF( radius, radius ), scaledPosition + QPointF( radius, radius ) );
        painter.drawLine( scaledPosition - QPointF( radius, -radius ), scaledPosition + QPointF( radius, -radius ) );
    }

    // horizontal errors
    painter.setPen( oldPen );
    if( point.errorX().isValid() )
    {
        const RefPoint::PointF minus( point.errorX().globalMinus_*realZoomFactor_ );
        const RefPoint::PointF plus( point.errorX().globalPlus_*realZoomFactor_ );
        const RefPoint::PointF offset( RefPoint::PointF( minus.y() - plus.y(), plus.x() - minus.x() )*(0.5*pointSize/plus.length( minus ) ) );
        painter.drawLine( minus, plus );

        if( hoverMode == RefPoint::HoverErrorXMinus ) painter.setPen( hoverPen );
        painter.drawLine( minus - offset, minus + offset );

        if( hoverMode == RefPoint::HoverErrorXPlus ) painter.setPen( hoverPen );
        else painter.setPen( oldPen );
        painter.drawLine( plus - offset, plus + offset );
    }

    // vertical errors
    painter.setPen( oldPen );
    if( point.errorY().isValid() )
    {
        const RefPoint::PointF minus( point.errorY().globalMinus_*realZoomFactor_ );
        const RefPoint::PointF plus( point.errorY().globalPlus_*realZoomFactor_ );
        const RefPoint::PointF offset( RefPoint::PointF( minus.y() - plus.y(), plus.x() - minus.x() )*(0.5*pointSize/plus.length( minus ) ) );
        painter.drawLine( minus, plus );

        if( hoverMode == RefPoint::HoverErrorYMinus ) painter.setPen( hoverPen );
        painter.drawLine( minus - offset, minus + offset );

        if( hoverMode == RefPoint::HoverErrorYPlus ) painter.setPen( hoverPen );
        else painter.setPen( oldPen );
        painter.drawLine( plus - offset, plus + offset );
    }

    // vertical systematic errors
    painter.setPen( oldPen );
    if( point.errorSystY().isValid() )
    {
        const RefPoint::PointF minus( point.errorSystY().globalMinus_*realZoomFactor_ );
        const RefPoint::PointF plus( point.errorSystY().globalPlus_*realZoomFactor_ );
        const RefPoint::PointF offset( RefPoint::PointF( minus.y() - plus.y(), plus.x() - minus.x() )*(1.5*pointSize/plus.length( minus ) ) );
        painter.drawLine( minus - offset, plus - offset );
        painter.drawLine( minus + offset, plus + offset );

        if( hoverMode == RefPoint::HoverErrorSystYMinus ) painter.setPen( hoverPen );
        painter.drawLine( minus - offset, minus + offset );

        if( hoverMode == RefPoint::HoverErrorSystYPlus ) painter.setPen( hoverPen );
        else painter.setPen( oldPen );

        painter.drawLine( plus - offset, plus + offset );

    }

    return;

}

//________________________________________________________________________
RefPoint DisplayWidget::_findPoint( const QPointF& position, RefPoint::HoverMode& hoverMode ) const
{

    // apply zoom factor
    const RefPoint::PointF scaled( position*(1.0/realZoomFactor_) );
    const double pointSize( pointSize_*realZoomFactor_ );
    FindPointFTor findPointFTor( scaled, pointSize );

    // loop over points and find matching
    for( RefPoint::Set::const_iterator iter = points_.begin(); iter != points_.end(); ++iter )
    {

        if( ( hoverMode = findPointFTor.hoverMode( *iter ) ) != RefPoint::NoHover )
        { return findPointFTor.out(); }

    }

    // update mode and return
    if( findPointFTor.found() )
    {

        hoverMode = RefPoint::HoverPoint;
        return findPointFTor.out();

    }

    if( showInactivePointsAction().isChecked() )
    {

        // try find matching inactive points
        // loop over datasets
        for( int index = 0; index < (int) dataSets_.size(); index++ )
        {

            // skip current dataset
            if( index == dataSetIndex_ ) continue;

            // try find matching point
            const RefPoint::List& points( dataSets_[index] );
            for( RefPoint::List::const_iterator iter = points.begin(); iter != points.end(); ++iter )
            {

                if( ( hoverMode = findPointFTor.hoverMode( *iter ) ) != RefPoint::NoHover )
                { return findPointFTor.out(); }

            }

        }

        // update mode and return
        if( findPointFTor.found() )
        {

            hoverMode = RefPoint::HoverPoint;
            return findPointFTor.out();

        }

    }

    // try find matching reference point
    for( RefPoint::Set::const_iterator iter = refPoints_.begin(); iter != refPoints_.end(); ++iter )
    {
        if( showRefPointsAction().isChecked() || !refPointsValid_ || selectedPoints_.find( *iter ) != selectedPoints_.end() )
        { findPointFTor.hoverMode( *iter ); }
    }

    // update mode and return
    if( findPointFTor.found() )
    {

        hoverMode = RefPoint::HoverRefPoint;
        return findPointFTor.out();

    }

    // nothing found. Just return
    hoverMode = RefPoint::NoHover;
    return RefPoint();

}

//________________________________________________________________________
void DisplayWidget::_toggleShowRefPoints( bool value )
{
    XmlOptions::get().set<bool>( "SHOW_REF_POINTS", value );
    XmlOptions::write();
    update();
}

//________________________________________________________________________
void DisplayWidget::_toggleShowInactivePoints( bool value )
{
    XmlOptions::get().set<bool>( "SHOW_INACTIVE_POINTS", value );
    XmlOptions::write();
    update();
}

//________________________________________________________________________
void DisplayWidget::_toggleShowScanLines( bool value )
{
    XmlOptions::get().set<bool>( "SHOW_SCAN_LINES", value );
    XmlOptions::write();
    update();
}

//________________________________________________________________________
void DisplayWidget::_addPoint( void )
{
    Debug::Throw( "DisplayWidget::_addPoint.\n" );
    emit pointAdded( QPointF( position_ )*(1.0/realZoomFactor_) );
}

//________________________________________________________________________
void DisplayWidget::_pastePoint( void )
{
    Debug::Throw( "DisplayWidget::_pastePoint.\n" );
    emit pointPasted( QPointF( position_ )*(1.0/realZoomFactor_) );
}

//______________________________________
void DisplayWidget::_removeSelection( void )
{ emit selectionRemoved( selectedPoints_.toList() ); }

//______________________________________
void DisplayWidget::_updateConfiguration( void )
{
    Debug::Throw( "DisplayWidget::_updateConfiguration.\n" );

    pointSize_ = 0.5*XmlOptions::get().get<int>( "POINT_SIZE" );
    penWidth_ = XmlOptions::get().get<int>( "PEN_SIZE" );

    inactivePointColor_ = QColor( XmlOptions::get().get<BASE::Color>( "INACTIVE_POINT_COLOR" ) );
    pointColor_ = QColor( XmlOptions::get().get<BASE::Color>( "POINT_COLOR" ) );
    refPointColor_ = QColor( XmlOptions::get().get<BASE::Color>( "REF_POINT_COLOR" ) );
    selectedPointColor_ = QColor( XmlOptions::get().get<BASE::Color>( "SELECTED_POINT_COLOR" ) );
    hoverPointColor_ = QColor( XmlOptions::get().get<BASE::Color>( "HOVER_POINT_COLOR" ) );

    showInactivePointsAction().setChecked( XmlOptions::get().get<bool>( "SHOW_INACTIVE_POINTS" ) );
    showRefPointsAction().setChecked( XmlOptions::get().get<bool>( "SHOW_REF_POINTS" ) );
    showScanLinesAction().setChecked( XmlOptions::get().get<bool>( "SHOW_SCAN_LINES" ) );

    // repaint
    update();

    return;
}

//________________________________________________________________________
void DisplayWidget::_adjustZoomFactor( void )
{

    Debug::Throw( "DisplayWidget::_adjustZoomFactor.\n" );

    // get parent widget
    QWidget* parent( parentWidget() );
    QAbstractScrollArea* parentFrame( qobject_cast<QAbstractScrollArea*>(parent->parentWidget()) );
    Q_CHECK_PTR( parentFrame );

    QSize size( parentFrame->size() );
    size.setWidth( size.width() - 2.0*parentFrame->frameWidth() );
    size.setHeight( size.height() - 2.0*parentFrame->frameWidth() );

    // add extra margin
    const int margin(2);
    size.setWidth( size.width() - 2.0*margin );
    size.setHeight( size.height() - 2.0*margin );

    ZoomWidget::Mode zoomMode( zoomMode_ );
    if( zoomMode == ZoomWidget::FitPage )
    {
        const double scaleWidth( double( size.width() )/double( pixmap_.width()) );
        const double scaleHeight( double( size.height() )/double( pixmap_.height()) );
        zoomMode = scaleWidth < scaleHeight ? ZoomWidget::FitWidth : ZoomWidget::FitHeight;
    }

    switch( zoomMode )
    {

        case ZoomWidget::FitWidth:
        zoomFactor_ = double(size.width())/double( pixmap_.width());

        // account for scrollbar size
        if( pixmap_.height()*zoomFactor_ > size.height() )
        { zoomFactor_ -= double( parentFrame->verticalScrollBar()->width() )/double( pixmap_.width()); }
        break;

        case ZoomWidget::FitHeight:
        zoomFactor_ = double(size.height())/double( pixmap_.height() );

        // account for scrollbar size
        if( pixmap_.width()*zoomFactor_ > size.width() )
        { zoomFactor_ -= double( parentFrame->horizontalScrollBar()->height() )/double( pixmap_.height()); }

        case ZoomWidget::FitPage:
        {}

        default: break;

    }

    return;

}

//________________________________________________________________________
RefPoint::HoverMode DisplayWidget::FindPointFTor::hoverMode( const RefPoint& point )
{

    // compare to error X ticks
    if( point.hasErrorX() )
    {

        // initialize vectX
        if( !vectXInitialized_ )
        {
            vectX_ = RefPoint::PointF( point.errorX().globalPlus_ - point.errorX().globalMinus_ ).normalized();
            vectXInitialized_ = true;
        }

        const RefPoint::PointF local( position_.rotated( vectX_ ) );
        const RefPoint::PointF localPlus( point.errorX().globalPlus_.rotated( vectX_ ) );
        if( std::abs( local.x() - localPlus.x() ) <= 1 && std::abs( local.y() - localPlus.y() ) <= 0.5*pointSize_ )
        {

            found_ = true;
            out_ = point;
            return RefPoint::HoverErrorXPlus;

        }

        const RefPoint::PointF localMinus( point.errorX().globalMinus_.rotated( vectX_ ) );
        if( std::abs( local.x() - localMinus.x() ) <= 1 && std::abs( local.y() - localMinus.y() ) <= 0.5*pointSize_ )
        {
            found_ = true;
            out_ = point;
            return RefPoint::HoverErrorXMinus;
        }
    }

    // compare to error Y ticks
    if( point.hasErrorY() )
    {

        // initialize vectY
        if( !vectYInitialized_ )
        {
            vectY_ = RefPoint::PointF( point.errorY().globalPlus_ - point.errorY().globalMinus_ ).normalized();
            vectYInitialized_ = true;
        }

        const RefPoint::PointF local( position_.rotated( vectY_ ) );
        const RefPoint::PointF localPlus( point.errorY().globalPlus_.rotated( vectY_ ) );
        if( std::abs( local.x() - localPlus.x() ) <= 1 && std::abs( local.y() - localPlus.y() ) <= 0.5*pointSize_ )
        {
            found_ = true;
            out_ = point;
            return RefPoint::HoverErrorYPlus;
        }

        const RefPoint::PointF localMinus( point.errorY().globalMinus_.rotated( vectY_ ) );
        if( std::abs( local.x() - localMinus.x() ) <= 1 && std::abs( local.y() - localMinus.y() ) <= 0.5*pointSize_ )
        {
            found_ = true;
            out_ = point;
            return RefPoint::HoverErrorYMinus;
        }
    }

    // compare to systematic error Y ticks
    if( point.hasErrorSystY() )
    {

        // initialize vectY
        if( !vectYInitialized_ )
        {
            vectY_ = RefPoint::PointF( point.errorSystY().globalPlus_ - point.errorSystY().globalMinus_ ).normalized();
            vectYInitialized_ = true;
        }

        const RefPoint::PointF local( position_.rotated( vectY_ ) );
        const RefPoint::PointF localPlus( point.errorSystY().globalPlus_.rotated( vectY_ ) );
        if( std::abs( local.x() - localPlus.x() ) <= 1 && std::abs( local.y() - localPlus.y() ) <= 1.5*pointSize_ )
        {
            found_ = true;
            out_ = point;
            return RefPoint::HoverErrorSystYPlus;
        }

        const RefPoint::PointF localMinus( point.errorSystY().globalMinus_.rotated( vectY_ ) );
        if( std::abs( local.x() - localMinus.x() ) <= 1 && std::abs( local.y() - localMinus.y() ) <= 1.5*pointSize_ )
        {
            found_ = true;
            out_ = point;
            return RefPoint::HoverErrorSystYMinus;
        }
    }

    // compare to center
    if( (position_ - point.global()).manhattanLength() > pointSize_ ) return RefPoint::NoHover;

    // full length
    const double distance( position_.length( point.global() ) );
    if( distance > pointSize_ ) return RefPoint::NoHover;

    // compare to minDistance
    if( distance < minDistance_ || minDistance_ < 0 )
    {
        minDistance_ = distance;
        out_ = point;
        found_ = true;
    }

    return RefPoint::NoHover;

}
