#ifndef DisplayWidget_h
#define DisplayWidget_h

// $Id: DisplayWidget.h,v 1.38 2013/02/14 09:02:46 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "Counter.h"
#include "File.h"
#include "RefPoint.h"
#include "ZoomWidget.h"

#include <QDialog>
#include <QMenu>
#include <QPaintEvent>
#include <QPixmap>
#include <QWidget>

#include <QBasicTimer>

class DisplayWidget: public QWidget, public Counter
{

    Q_OBJECT

    public:

    //! constructor
    DisplayWidget( QWidget* = 0 );

    //! destructor
    virtual ~DisplayWidget( void )
    {}

    //! set image
    void setImage( const File& );

    //! file
    const File& file( void ) const
    { return file_; }

    //! clear
    void clear( void )
    {
        file_ = File();
        refPoints_.clear();
        points_.clear();
        selectedPoints_.clear();
        hoverMode_ = RefPoint::NoHover;
        dragMode_ = RefPoint::NoDrag;
        update();
    }

    //! ref points
    void setRefPoints( const RefPoint::Set& value )
    {
        refPoints_ = value;
        update();
    }

    //! ref points validity
    void setRefPointsValid( bool value )
    { refPointsValid_ = value; }

    //! collection
    void setCollection( const RefPoint::Collection& collection )
    { dataSets_ = collection; }

    //! index
    void setDataSetIndex( int value )
    { dataSetIndex_ = value; }

    //! points
    void setPoints( const RefPoint::Set& value )
    {
        points_ = value;

        // update hover point
        _updateHoverPoint();

        // update drag point
        if( dragMode_ != RefPoint::NoDrag && points_.find( dragPoint_ ) == points_.end() )
        { dragMode_ = RefPoint::NoDrag; }

        update();
    }

    //! restore drag mode
    void setDragMode( RefPoint::DragMode dragMode )
    { dragMode_ = dragMode; }

    //! add single point
    void addPoint( const RefPoint& value )
    {
        points_.insert( value );
        _updateHoverPoint();
        update();
    }

    //! selected points
    void clearSelection( void )
    {
        selectedPoints_.clear();
        update();
    }

    //! selected points
    void setSelection( const RefPoint& value )
    {
        selectedPoints_.clear();
        selectedPoints_.insert( value );
        update();
    }

    //! selected points
    void setSelection( const RefPoint::List& value )
    {
        selectedPoints_.clear();
        foreach( const RefPoint& point, value )
        { selectedPoints_.insert( point ); }
        update();
    }

    //! selected points
    void setSelection( const RefPoint::Set& value )
    {
        selectedPoints_ = value;
        update();
    }

    //!@name actions
    //@{

    QAction& showRefPointsAction( void ) const
    { return *showRefPointsAction_; }

    QAction& showInactivePointsAction( void ) const
    { return *showInactivePointsAction_; }

    QAction& showScanLinesAction( void ) const
    { return *showScanLinesAction_; }

    QAction& addPointAction( void ) const
    { return *addPointAction_; }

    QAction& removePointAction( void ) const
    { return *removePointAction_; }

    QAction& copyPointAction( void ) const
    { return *copyPointAction_; }

    QAction& pastePointAction( void ) const
    { return *pastePointAction_; }

    QAction& editErrorsAction( void ) const
    { return *editErrorsAction_; }

    QAction& defaultErrorsAction( void ) const
    { return *defaultErrorsAction_; }

    QAction& defaultErrorsSystAction( void ) const
    { return *defaultErrorsSystAction_; }

    //@}

    //! event handling
    virtual bool eventFilter( QObject*, QEvent* );

    //! true if current position is valid
    bool positionIsValid( void );

    //! hover point
    const RefPoint& hoverPoint( void ) const
    { return hoverPoint_; }

    //! hover point
    void setHoverPoint( const RefPoint& value )
    { hoverPoint_ = value; }

    //! hover mode
    RefPoint::HoverMode hoverMode( void ) const
    { return hoverMode_; }

    //! drag point
    const RefPoint& dragPoint( void ) const
    { return dragPoint_; }

    //! drag point
    void setDragPoint( const RefPoint& value )
    { dragPoint_ = value; }

    //! drag mode
    RefPoint::DragMode dragMode( void ) const
    { return dragMode_; }

    //! real zoom factor
    double realZoomFactor( void ) const
    { return realZoomFactor_; }

    signals:

    //! mouse position changed
    void pointerMoved( QPoint, RefPoint::HoverMode );

    //! mouse left widget
    void widgetEntered( void );

    //! mouse left widget
    void widgetLeft( void );

    //! real factor
    void realZoomFactorChanged( double );

    //! point selected
    void pointSelected( void );

    //! point added
    void pointAdded( QPointF );

    //! point pasted
    void pointPasted( QPointF );

    //! point dragged
    void pointDragged( QPointF, RefPoint::DragMode );

    //! remove selected markers
    void selectionRemoved( const RefPoint::List& );

    public slots:

    //! zoom modified function
    void zoomModified( ZoomWidget::Mode, double );

    //! update position
    void updatePosition( void );

    //! update mouse position
    bool setPosition( QPoint );

    protected:

    //!@name event handling
    //@{
    virtual bool event( QEvent* );
    virtual void paintEvent( QPaintEvent* );
    virtual void timerEvent( QTimerEvent* );
    virtual void mouseMoveEvent( QMouseEvent* );
    virtual void mousePressEvent( QMouseEvent* );
    virtual void mouseReleaseEvent( QMouseEvent* );
    virtual void contextMenuEvent( QContextMenuEvent* );
    //@}

    //! adjust zoom factor
    void _adjustZoomFactor( void );

    //! scaled pixmap
    void _updateScaledPixmap( void );

    //! update hover point
    bool _updateHoverPoint( void );

    //! draw symbol
    void _drawSymbol( QPainter&, const RefPoint&, const RefPoint::HoverMode& = RefPoint::NoHover ) const;

    //! find index of nearby point, if any
    RefPoint _findPoint( const QPointF&, RefPoint::HoverMode& ) const;

    class FindPointFTor
    {

        public:

        //! constructor
        FindPointFTor( RefPoint::PointF position, const double& pointSize ):
            position_( position ),
            pointSize_( pointSize ),
            minDistance_( -1 ),
            found_( false ),
            vectXInitialized_( false ),
            vectYInitialized_( false )
        {}

        //! hover mode for given ref point
        RefPoint::HoverMode hoverMode( const RefPoint& );

        //! true if one point was found
        bool found( void ) const
        { return found_; }

        //! found point
        const RefPoint& out( void )
        { return out_; }

        private:

        //! predicted position (scaled)
        RefPoint::PointF position_;

        //! point size (scaled)
        double pointSize_;

        //! min distance to point center
        double minDistance_;

        //! true if point is found
        bool found_;

        //! found point
        RefPoint out_;

        //! true if X vector is initialized
        bool vectXInitialized_;

        //! true if Y vector is initialized
        bool vectYInitialized_;

        //! X vector
        RefPoint::PointF vectX_;

        //! Y vector
        RefPoint::PointF vectY_;

    };

    protected slots:

    //! toggle show ref point
    void _toggleShowRefPoints( bool );

    //! toggle show ref point
    void _toggleShowInactivePoints( bool );

    //! toggle show scan lines
    void _toggleShowScanLines( bool );

    //! add point
    void _addPoint( void );

    //! paste point
    void _pastePoint( void );

    //! remove selection
    void _removeSelection( void );

    private slots:

    //! update configuration
    void _updateConfiguration( void );

    private:

    //! install actions
    void _installActions( void );

    //! image
    File file_;

    //! pixmap
    QPixmap pixmap_;

    //! pixmap
    QPixmap scaledPixmap_;

    //! current position
    QPoint position_;

    //! click position
    QPoint clickPosition_;

    //! reference points
    RefPoint::Set refPoints_;

    //! ref point validity
    bool refPointsValid_;

    //! collection
    RefPoint::Collection dataSets_;

    //! collection index
    int dataSetIndex_;

    //! points
    RefPoint::Set points_;

    //! selected points
    RefPoint::Set selectedPoints_;

    //!@name options
    //@{

    double pointSize_;
    double penWidth_;
    QColor inactivePointColor_;
    QColor pointColor_;
    QColor refPointColor_;
    QColor selectedPointColor_;
    QColor hoverPointColor_;

    //@}

    //!@name actions
    //@{

    QAction* showRefPointsAction_;
    QAction* showInactivePointsAction_;
    QAction* showScanLinesAction_;
    QAction* addPointAction_;
    QAction* removePointAction_;
    QAction* copyPointAction_;
    QAction* pastePointAction_;
    QAction* editErrorsAction_;
    QAction* defaultErrorsAction_;
    QAction* defaultErrorsSystAction_;

    //@}

    //! resize timer
    QBasicTimer resizeTimer_;

    //!@name zoom
    //@{
    ZoomWidget::Mode zoomMode_;
    double zoomFactor_;
    double realZoomFactor_;
    //@}

    //!@name hover
    //@{
    RefPoint::HoverMode hoverMode_;
    RefPoint hoverPoint_;
    //@}

    //!@name drag
    //@{
    RefPoint::DragMode dragMode_;
    RefPoint dragPoint_;
    //@}

};

#endif
