// $Id: ImageRecord.cpp,v 1.11 2012/12/09 22:20:53 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "ImageRecord.h"
#include "XmlFileRecord.h"
#include "XmlDef.h"

//_____________________________________________________________________________
ImageRecord::ImageRecord( const File& file, int version ):
    Counter( "ImageRecord" ),
    file_( file ),
    version_( version ),
    dataSetIndex_( 0 ),
    flags_( 0 )
{
    // make sure there is at least one data set
    dataSets_ << RefPoint::List();
}

//_____________________________________________________________________________
ImageRecord::ImageRecord( const QDomElement& element ):
    Counter( "ImageRecord" ),
    version_( 0 ),
    dataSetIndex_( 0 ),
    flags_( 0 )
{
    Debug::Throw( "ImageRecord::ImageRecord.\n" );

    QDomNamedNodeMap attributes( element.attributes() );
    for( unsigned int i=0; i<attributes.length(); i++ )
    {
        QDomAttr attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;

        QString name( attribute.name() );
        QString value( attribute.value() );
        if( name == XML::FLAGS ) flags_ = value.toInt();
        else if( name == XML::INDEX ) dataSetIndex_ = value.toInt();
        else if( name == XML::IMAGE_VERSION ) version_ = value.toUInt();
        else Debug::Throw(0) << "ImageRecord::ImageRecord - unrecognized entry attribute: \"" << name << "\"" << endl;

    }

    // parse children elements
    for(QDomNode childNode = element.firstChild(); !childNode.isNull(); childNode = childNode.nextSibling() )
    {
        QDomElement childElement = childNode.toElement();
        if( childElement.isNull() ) continue;

        QString tagName( childElement.tagName() );
        if( tagName == FILERECORD::XML::RECORD ) file_ = XmlFileRecord( childElement );

        // point lists
        else if( tagName == XML::REF_POINTS ) refPoints_ = RefPoint::List( childElement );
        else if( tagName == XML::DATA_SET ) _addDataSet( RefPoint::List( childElement ) );
        else Debug::Throw(0) << "XmlFileRecord::XmlFileRecord - unrecognized child " << childElement.tagName() << endl;
    }

    // make sure there is at least one data set
    if( dataSets_.empty() )
    { dataSets_ << RefPoint::List(); }
}

//_______________________________________________
QDomElement ImageRecord::domElement( QDomDocument& parent ) const
{

    Debug::Throw( "ImageRecord::domElement.\n" );

    // make sure dataSetIndex points to a valid data set
    int localIndex( dataSetIndex_ );
    if( localIndex < 0 || localIndex >= (int)dataSets_.size() || dataSets_[localIndex].empty() )
    { localIndex = 0; }

    QDomElement out( parent.createElement( XML::IMAGE_RECORD ) );
    out.setAttribute( XML::INDEX, QString::number( localIndex ) );
    out.setAttribute( XML::FLAGS, QString::number( flags_ ) );
    if( version_ > 0 ) out.setAttribute( XML::IMAGE_VERSION, QString::number( version_ ) );

    out.appendChild( XmlFileRecord( file_ ).domElement( parent ) );
    out.appendChild( refPoints_.domElement( XML::REF_POINTS, parent ) );

    for( RefPoint::Collection::const_iterator iter = dataSets_.begin(); iter != dataSets_.end(); ++iter )
    {
        if( !iter->empty() )
        { out.appendChild( iter->domElement( XML::DATA_SET, parent ) ); }
    }

    return out;

}

//_______________________________________________
ImageRecord::List::List( const QDomElement& element ):
    Counter( "ImageRecord::List" )
{

    // parse children elements
    for(QDomNode childNode = element.firstChild(); !childNode.isNull(); childNode = childNode.nextSibling() )
    {
        QDomElement element = childNode.toElement();
        if( element.isNull() ) continue;

        // special options
        if( element.tagName() == XML::IMAGE_RECORD )
        {

            ImageRecord record( element );
            if( !record.file().file().isEmpty() ) append( record );
            else Debug::Throw(0, "ImageRecord::List::List - attend to add empty record. Discarded.\n" );

        } else Debug::Throw() << "ImageRecord::List::List - unrecognized tag " << element.tagName() << endl;
    }

}

//_______________________________________________
QDomElement ImageRecord::List::domElement( QDomDocument& parent ) const
{

    Debug::Throw( "ImageRecord::List::domElement.\n" );

    // create main element
    QDomElement element = parent.appendChild( parent.createElement( XML::IMAGE_RECORD_LIST ) ).toElement();

    // loop over records
    foreach( const ImageRecord& record, *this )
    {

        Debug::Throw() << "ImageRecord::List::domElement - " << record.file();
        if( record.file().file().isEmpty() )
        {
            Debug::Throw(0, "ImageRecord::List::domElement - attend to write empty record. Discarded.\n" );
            continue;
        }

        element.appendChild( record.domElement( parent ) );
    }

    return element;

}
