// $Id: ImageRecordConfiguration.cpp,v 1.10 2013/03/12 08:47:46 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "ImageRecordConfiguration.h"

#include "BaseIcons.h"
#include "ContextMenu.h"
#include "Debug.h"
#include "ImageRecordList.h"
#include "GridLayout.h"
#include "IconEngine.h"
#include "OptionCheckBox.h"
#include "OptionSpinBox.h"
#include "QuestionDialog.h"
#include "TreeView.h"

#include <QGroupBox>
#include <QLabel>
#include <QLayout>
#include <QShortcut>

//___________________________________________
ImageRecordConfiguration::ImageRecordConfiguration( QWidget* parent, ImageRecordList& recent_files ):
    QWidget( parent ),
    Counter( "ImageRecordConfiguration" ),
    recentFiles_( &recent_files )
{
    Debug::Throw( "ImageRecordConfiguration::ImageRecordConfiguration.\n" );

    QVBoxLayout* layout = new QVBoxLayout();
    setLayout( layout );
    layout->setMargin(0);
    layout->setSpacing( 5 );

    // group box for additional options
    QGroupBox* box;
    layout->addWidget( box = new QGroupBox( tr( "Options" ), this ) );

    QVBoxLayout* vLayout = new QVBoxLayout();
    vLayout->setMargin(5);
    vLayout->setSpacing(5);
    box->setLayout( vLayout );

    QHBoxLayout* hLayout = new QHBoxLayout();
    hLayout->setSpacing(5);
    hLayout->setMargin(0);
    vLayout->addLayout( hLayout );


    // previous file history size
    OptionSpinBox *spinbox;
    hLayout->addWidget( new QLabel( tr( "Recent files history size (0 means <i>unlimited</i>):" ), box ) );
    hLayout->addWidget( spinbox = new OptionSpinBox( box, "DB_SIZE" ) );
    hLayout->addStretch(1);
    spinbox->setMinimum( 0 );
    spinbox->setMaximum( 100 );
    addOptionWidget( spinbox );
    spinbox->setToolTip( tr( "Number of previously opened files to appear in the Open Previous menu" ) );

    // sort previous files by date
    OptionCheckBox* checkbox;
    vLayout->addWidget( checkbox = new OptionCheckBox( tr( "Sort recent files by date" ), box, "SORT_FILES_BY_DATE" ) );
    checkbox->setToolTip( tr( "Sort files by date rather than name in Open Previous menu." ) );
    addOptionWidget( checkbox );

    // group box for file list
    layout->addWidget( box = new QGroupBox( tr( "Files" ), this ), 1 );
    box->setLayout( new QHBoxLayout() );
    box->layout()->setSpacing(5);
    box->layout()->setMargin(5);

    box->layout()->addWidget( list_ = new TreeView( box ) );
    _list().setModel( &model_ );
    _list().sortByColumn( XmlOptions::get().get<bool>( "SORT_FILES_BY_DATE" ) ? ImageRecordModel::TIME:ImageRecordModel::FILE );
    _list().setSelectionMode( QAbstractItemView::ContiguousSelection );
    _list().setOptionName( "RECENT_FILES_CONFIGURATION_LIST" );

    connect( _list().selectionModel(), SIGNAL( selectionChanged(const QItemSelection &, const QItemSelection &) ), SLOT( _updateButtons() ) );

    vLayout = new QVBoxLayout();
    vLayout->setSpacing(5);
    vLayout->setMargin(0);
    box->layout()->addItem( vLayout );

    // context menu
    QMenu* menu( new ContextMenu( &_list() ) );

    // clean
    vLayout->addWidget( cleanButton_ = new QPushButton( tr( "Clean" ), box ) );
    cleanButton_->setToolTip( tr( "Remove invalid files" ) );
    cleanButton_->setIcon( IconEngine::get( ICONS::DELETE ) );
    connect( cleanButton_, SIGNAL( clicked() ), SLOT( _clean() ) );

    addAction( cleanAction_ = new QAction( IconEngine::get( ICONS::DELETE ), tr( "Clean" ), this ) );
    connect( cleanAction_, SIGNAL( triggered() ), SLOT( _clean() ) );
    menu->addAction( cleanAction_ );

    // remove
    vLayout->addWidget( removeButton_ = new QPushButton( tr( "Remove" ), box ) );
    removeButton_->setIcon( IconEngine::get( ICONS::REMOVE ) );
    removeButton_->setToolTip( tr( "Remove selected files" ) );
    connect( removeButton_, SIGNAL( clicked() ), SLOT( _remove() ) );

    addAction( removeAction_ = new QAction( IconEngine::get( ICONS::REMOVE ), tr( "Remove" ), this ) );
    connect( removeAction_, SIGNAL( triggered() ), SLOT( _remove() ) );
    removeAction_->setShortcut( QKeySequence::Delete );
    menu->addAction( removeAction_ );
    _list().addAction( removeAction_ );

    // reload
    vLayout->addWidget( reloadButton_ = new QPushButton( tr( "Reload" ), box ) );
    reloadButton_->setToolTip( tr( "Reload file list" ) );
    reloadButton_->setIcon( IconEngine::get( ICONS::RELOAD ) );
    connect( reloadButton_, SIGNAL( clicked() ), SLOT( _reload() ) );

    addAction( reloadAction_ = new QAction( IconEngine::get( ICONS::RELOAD ), tr( "Reload" ), this ) );
    connect( reloadAction_, SIGNAL( triggered() ), SLOT( _reload() ) );
    menu->addAction( reloadAction_ );

    vLayout->addStretch( 1 );

    // connections
    _reloadButton().setEnabled( false );

}

//__________________________________________________________________________
void ImageRecordConfiguration::read( void )
{
    Debug::Throw( "ImageRecordConfiguration::read.\n" );
    OptionWidgetList::read();

    _reload();
    _list().resizeColumns();

}

//__________________________________________________________________________
void ImageRecordConfiguration::write( void ) const
{
    Debug::Throw( "ImageRecordConfiguration::write.\n" );
    OptionWidgetList::write();

    // put model contents into file list
    _recentFiles().set( _model().get() );

}

//__________________________________________________________________________
void ImageRecordConfiguration::_updateButtons( void )
{

    Debug::Throw( "ImageRecordConfiguration::_updateButtons.\n" );
    QList<QModelIndex> selection( _list().selectionModel()->selectedRows() );
    _removeButton().setEnabled( !selection.empty() );

}

//__________________________________________________________________________
void ImageRecordConfiguration::_remove( void )
{

    Debug::Throw( "ImageRecordConfiguration::_remove.\n" );
    QList<QModelIndex> selection( _list().selectionModel()->selectedRows() );
    if( selection.isEmpty() ) return;

    // ask confirmation
    if( !QuestionDialog( this, tr( "Remove selected files from list ?" ) ).centerOnParent().exec() ) return;

    _model().remove( _model().get( selection ) );
    _list().selectionModel()->clear();

    _reloadButton().setEnabled( true );

    // records
    ImageRecordModel::List records( _model().get() );
    _cleanButton().setEnabled( std::find_if( records.begin(), records.end(), ImageRecord::InvalidFTor() ) != records.end() );

}

//__________________________________________________________________________
void ImageRecordConfiguration::_clean( void )
{
    Debug::Throw( "ImageRecordConfiguration::_clean.\n" );

    // remove invalid files
    ImageRecordModel::List records( _model().get() );
    records.erase(
        std::remove_if( records.begin(), records.end(), ImageRecord::InvalidFTor() ),
        records.end() );

    _model().set( records );

    _reloadButton().setEnabled( true );
    _cleanButton().setEnabled( false );
    _updateButtons();

}

//__________________________________________________________________________
void ImageRecordConfiguration::_reload( void )
{

    Debug::Throw( "ImageRecordConfiguration::_reload.\n" );
    ImageRecord::List records( _recentFiles().records() );
    _model().set( records );
    _list().selectionModel()->clear();

    _reloadButton().setEnabled( false );
    _cleanButton().setEnabled( std::find_if( records.begin(), records.end(), ImageRecord::InvalidFTor() ) != records.end() );
    _updateButtons();

}
