// $Id: ImageRecordList.cpp,v 1.23 2013/02/14 09:02:46 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*******************************************************************************/

#include "ImageRecordList.h"

#include "Debug.h"
#include "File.h"
#include "Str.h"
#include "Singleton.h"
#include "XmlDef.h"
#include "XmlDocument.h"
#include "XmlOptions.h"

#include <QApplication>
#include <algorithm>

//_______________________________________________
ImageRecordList::ImageRecordList( QObject* parent ):
    QObject( parent ),
    Counter( "ImageRecordList" ),
    maxSize_( -1 ),
    check_( true ),
    cleanEnabled_( false ),
    thread_( this )
{
    connect( Singleton::get().application(), SIGNAL( configurationChanged() ), SLOT( _updateConfiguration() ) );
    connect( qApp, SIGNAL( aboutToQuit() ), SLOT( _saveConfiguration() ) );
    _updateConfiguration();

    // thread configuration
    thread_.setCheckDuplicates( false );
    connect( &thread_, SIGNAL( recordsAvailable( const FileRecord::List&, bool ) ), this, SLOT( _processRecords( const FileRecord::List&, bool ) ) );

}

//_______________________________________________
bool ImageRecordList::read( File file )
{

    Debug::Throw( "ImageRecordList::_read.\n" );
    if( file.isEmpty() ) file = dbFile_;
    if( dbFile_.isEmpty() ) return false;

    // parse the file
    XmlDocument document;
    {
        QFile qfile( file );
        if( !document.setContent( &qfile ) )
        {
            Debug::Throw() << document.error() << endl;
            return false;
        }
    }

    // look for relevant element
    QDomNodeList topNodes = document.elementsByTagName( XML::IMAGE_RECORD_LIST );
    if( topNodes.isEmpty() ) return false;
    for(QDomNode node = topNodes.at(0).firstChild(); !node.isNull(); node = node.nextSibling() )
    {
        QDomElement element = node.toElement();
        if( element.isNull() ) continue;

        // special options
        if( element.tagName() == XML::IMAGE_RECORD )
        {

            ImageRecord record( element );
            if( !record.file().file().isEmpty() ) _add( record, true, false );
            else Debug::Throw(0, "ImageRecordList::_read - attend to add empty record. Discarded.\n" );

        } else Debug::Throw() << "ImageRecordList::_read - unrecognized tag " << element.tagName() << endl;
    }

    emit contentsChanged();

    return true;
}

//_______________________________________________
bool ImageRecordList::write( File file )
{
    Debug::Throw( "ImageRecordList::write.\n" );
    if( file.isEmpty() ) file = dbFile_;
    if( dbFile_.isEmpty() ) return false;

    // create document and read
    XmlDocument document;
    {
        QFile qtfile( file );
        document.setContent( &qtfile );
    }

    // append truncated list
    QDomElement top = _truncatedList( _records() ).domElement( document );
    document.replaceChild( top );
    {
        QFile qfile( file );
        if( !qfile.open( QIODevice::WriteOnly ) ) return false;
        qfile.write( document.toByteArray() );
    }

    return true;
}

//_______________________________________________
bool ImageRecordList::_setDBFile( const File& file )
{
    Debug::Throw() << "ImageRecordList::_setDBFile - file: " << file << endl;

    // check file
    if( dbFile_ == file && !_records().empty() ) return false;

    // store file and read
    dbFile_ = file;
    read();

    // make sure file is hidden (windows only)
    if( dbFile_.localName().startsWith('.') )
    { dbFile_.setHidden(); }

    return true;

}


//______________________________________
void ImageRecordList::_updateConfiguration( void )
{
    Debug::Throw( "ImageRecordList::_updateConfiguration.\n" );

    // DB file
    _setDBFile( File( XmlOptions::get().raw("RC_FILE") ) );
    _setMaxSize( XmlOptions::get().get<int>( "DB_SIZE" ) );
    return;

}

//______________________________________
void ImageRecordList::_saveConfiguration( void )
{
    Debug::Throw( "ImageRecordList::_saveConfiguration.\n" );
    write();
}

//_______________________________________________
bool ImageRecordList::contains( const File& file, int version ) const
{ return std::find_if( _records().begin(), _records().end(), ImageRecord::SameFileFTor( file, version ) ) != _records().end(); }

//_______________________________________________
void ImageRecordList::remove( const File& file, int version )
{
    Debug::Throw() << "ImageRecordList::remove - " << file << endl;
    _records().erase(std::remove_if( _records().begin(), _records().end(), ImageRecord::SameFileFTor( file, version ) ), _records().end() );
    return;
}

//___________________________________________________
void ImageRecordList::set( const ImageRecord::List& records )
{

    Debug::Throw( "ImageRecordList::set.\n" );
    _records() = records;
    emit contentsChanged();

}

//___________________________________________________
QList<File> ImageRecordList::files( void ) const
{
    Debug::Throw( "ImageRecordList::files.\n" );

    ImageRecord::List records( _truncatedList( _records() ) );

    QList<File> out;
    for( ImageRecord::List::const_iterator iter = records.begin(); iter != records.end(); ++iter )
    { out << iter->file().file(); }

    return out;
}

//_______________________________________________
ImageRecord ImageRecordList::lastValidFile( void )
{

    Debug::Throw( "ImageRecordList::lastValidFile.\n" );

    // sort list
    std::sort( _records().begin(), _records().end(), ImageRecord::FirstOpenFTor() );
    ImageRecord::ListIterator iter( _records() );
    iter.toBack();
    while( iter.hasPrevious() )
    {
        const ImageRecord& record( iter.previous() );
        if( !check_ || record.file().isValid() ) return record;
    }

    return ImageRecord( File("") );

}

//_______________________________________________
void ImageRecordList::checkValidFiles( void )
{
    Debug::Throw( "ImageRecordList::checkValidFiles.\n" );
    if( !check() ) return;
    if( thread_.isRunning() ) return;

    // not very optimal, but ...
    thread_.setRecords( _fileRecords().toList() );
    thread_.start();
}

//_______________________________________________
void ImageRecordList::_processRecords( const FileRecord::List& records, bool hasInvalidRecords )
{

    Debug::Throw( "ImageRecordList::_processRecords.\n" );

    // set file records validity
    ImageRecord::List& currentRecords( _records() );
    for( ImageRecord::List::iterator iter = currentRecords.begin(); iter != currentRecords.end(); ++iter )
    {

        FileRecord::List::const_iterator found = std::find_if(
            records.begin(),
            records.end(),
            FileRecord::SameFileFTor( iter->file().file() ) );
        if( found == records.end() )
        { continue; }

        iter->file().setValid( found->isValid() );
    }

    _setCleanEnabled( hasInvalidRecords );
    emit validFilesChecked();

}

//___________________________________________________
void ImageRecordList::clean( void )
{
    Debug::Throw( "ImageRecordList::clean" );

    if( !check() )
    {
        _records().clear();
        return;
    }

    // remove invalid files
    _records().erase(
        std::remove_if( _records().begin(), _records().end(), ImageRecord::InvalidFTor() ),
        _records().end() );

    return;
}

//___________________________________________________
void ImageRecordList::clear( void )
{
    Debug::Throw( "ImageRecordList::clear" );
    _records().clear();
    return;
}
//_______________________________________________
void ImageRecordList::_setMaxSize( const int& value )
{

    Debug::Throw( "ImageRecordList::_setMaxSize.\n" );
    maxSize_ = value;

    return;

}

//_______________________________________________
ImageRecord& ImageRecordList::_add(
    const ImageRecord& record,
    const bool& update,
    const bool& emitSignal )
{

    // do not add empty files
    Q_ASSERT( !record.file().file().isEmpty() );

    Debug::Throw() << "ImageRecordList::_add -"
        << " file: " << record.file().file()
        << " version " << record.version()
        << " update: " << update
        << endl;

    ImageRecord::List::iterator iter = std::find_if( _records().begin(), _records().end(), ImageRecord::SameFileFTor( record.file().file(), record.version() ) );
    if( iter != _records().end() )
    {

        if( update )
        {
            bool changed( false );
            if( iter->flags() != record.flags() )
            {
                changed = true;
                iter->setFlags( record.flags() );
            }

            if( iter->refPoints() != record.refPoints() )
            { changed = true; }

            if( iter->collection() != record.collection() )
            { changed = true; }

            if( iter->dataSetIndex() != record.dataSetIndex() )
            { changed = true; }

            // update none the less, because equal-to operator is incomplete
            iter->setRefPoints( record.refPoints() );
            iter->setCollection( record.collection() );
            iter->setDataSetIndex( record.dataSetIndex() );

            if( iter->file().time() != record.file().time() )
            {
                iter->file().setTime( qMax( iter->file().time(), record.file().time() ) );
                changed = true;
            }

            if( changed && emitSignal ) emit contentsChanged();
        }

        return *iter;

    } else {

        Debug::Throw() << "ImageRecordList::_add - record: " << record.file() << endl;

        _records() << record;

        if( emitSignal ) emit contentsChanged();
        return _records().back();

    }

}

//___________________________________________________________________
ImageRecord::List ImageRecordList::_truncatedList( ImageRecord::List records ) const
{
    // shrink list
    if( maxSize_ > 0 && int(records.size()) > maxSize_ )
    {

        std::sort( records.begin(), records.end(), ImageRecord::FirstOpenFTor() );
        if( check_ )
        {

            while( int(records.size()) > maxSize_ )
            {
                ImageRecord::List::iterator iter( std::find_if( records.begin(), records.end(), ImageRecord::InvalidFTor() ) );
                if( iter != records.end() ) records.erase( iter );
                else break;
            }

        }

        // remove oldest files
        while( int(records.size()) > maxSize_ )
        { records.removeFirst(); }

    }

    return records;
}
