// $Id: MainWindow.cpp,v 1.70 2013/03/22 09:41:40 hupereir Exp $
/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
****************************************************************************/

#include "MainWindow.h"

#include "Application.h"
#include "BaseStatusBar.h"
#include "Color.h"
#include "CustomToolBar.h"
#include "Debug.h"
#include "DisplayWidget.h"
#include "Icons.h"
#include "IconEngine.h"
#include "InformationDialog.h"
#include "Menu.h"
#include "NotificationWidget.h"
#include "QuestionDialog.h"
#include "RecentImagesMenu.h"
#include "RefPointDialog.h"
#include "RefPointErrorDialog.h"
#include "RefPointListWidget.h"
#include "RefPointPanel.h"
#include "Singleton.h"
#include "TreeView.h"
#include "UndoCommand.h"
#include "ZoomWidget.h"

#include <QTextStream>
#include <QDockWidget>
#include <QScrollBar>
#include <QShortcut>
#include <QSplitter>

//________________________________________________________________________________________
class ScrollArea: public QScrollArea, public Counter
{

    public:

    //! constructor
    ScrollArea( QWidget* parent = 0 ):
        QScrollArea( parent ),
        Counter( "QScrollArea" ),
        dragInProgress_( false ),
        xOrigin_(0),
        yOrigin_(0),
        oldCursor_( Qt::ArrowCursor )
    {}

    //! destructor
    virtual ~ScrollArea( void )
    {}

    //! event filter
    virtual bool eventFilter( QObject*, QEvent* );

    private:

    //! true if drag is in progress
    bool dragInProgress_;

    //! drag origin
    QPoint dragOrigin_;

    //! scrollbar position
    int xOrigin_;

    //! scrollbar position
    int yOrigin_;

    //! cursor
    Qt::CursorShape oldCursor_;

};

//________________________________________________________________________________________
bool ScrollArea::eventFilter( QObject* object, QEvent* event )
{

    switch( event->type() )
    {

        case QEvent::MouseMove:
        {

            QMouseEvent* mouseEvent( static_cast<QMouseEvent*>( event ) );
            if( !dragInProgress_ ) return false;

            // get position
            const QPoint position( mouseEvent->globalPos() );
            const QPoint delta( position - dragOrigin_ );
            if( horizontalScrollBar() ) horizontalScrollBar()->setValue( xOrigin_ - delta.x() );
            if( verticalScrollBar() ) verticalScrollBar()->setValue( yOrigin_ - delta.y() );

            return true;

        }

        case QEvent::MouseButtonPress:
        {
            QMouseEvent* mouseEvent( static_cast<QMouseEvent*>( event ) );

            // check button and modifier
            if( !( mouseEvent->button() == Qt::LeftButton && mouseEvent->modifiers() == Qt::ShiftModifier ) )
            { return false; }

            // setup drag
            dragInProgress_ = true;
            dragOrigin_ = mouseEvent->globalPos();
            xOrigin_ = (horizontalScrollBar() ? horizontalScrollBar()->value():0 );
            yOrigin_ = (verticalScrollBar() ? verticalScrollBar()->value():0 );

            // change cursor
            setCursor( Qt::ClosedHandCursor );
            oldCursor_ = widget()->cursor().shape();
            widget()->setCursor( Qt::ClosedHandCursor );
            return true;

        }

        case QEvent::MouseButtonRelease:
        {
            QMouseEvent* mouseEvent( static_cast<QMouseEvent*>( event ) );

            // check button and drag state
            if( !( mouseEvent->button() == Qt::LeftButton && dragInProgress_ ) )
            { return false; }

            // restor cursor
            unsetCursor();
            if( DisplayWidget* displayWidget = qobject_cast<DisplayWidget*>( widget() ) )
            {
                displayWidget->setCursor( oldCursor_ );
                displayWidget->setPosition( mouseEvent->pos() );
            }

            dragInProgress_ = false;
            return true;

        }

        default: break;

    }

    return QScrollArea::eventFilter( object, event );

}

//________________________________________________________________________________________
MainWindow::MainWindow( QWidget* parent ):
    BaseMainWindow( parent ),
    Counter( "MainWindow" ),
    wasHovered_( false )
{

    Debug::Throw( "MainWindow::MainWindow.\n" );
    setOptionName( "WINDOW" );

    // undo stack
    undoStack_ = new QUndoStack( this );

    // install actions
    _installActions();

    // main widget
    QWidget* main = new QWidget( this );
    setCentralWidget( main );

    // local layout
    QVBoxLayout *layout = new QVBoxLayout();
    layout->setMargin(0);
    layout->setSpacing(2);
    main->setLayout( layout );

    // top
    QWidget* top( new QWidget(main) );
    QVBoxLayout* topLayout = new QVBoxLayout();
    top->setLayout( topLayout );
    top->layout()->setMargin(0);
    top->layout()->setSpacing(2);
    layout->addWidget( top, 1 );

    // ref point panel
    layout->addWidget( refPointPanel_ = new RefPointPanel( main ), 0 );

    // horizontal splitter for display and list
    QSplitter* hSplitter( new QSplitter( top ) );
    topLayout->addWidget( hSplitter, 1 );
    hSplitter->setOrientation( Qt::Horizontal );

    // display widget
    scrollArea_ = new ScrollArea();
    displayWidget_ = new DisplayWidget();
    scrollArea_->setWidget( displayWidget_ );
    scrollArea_->setAlignment( Qt::AlignCenter );
    displayWidget_->installEventFilter( &refPointPanel_->zoomWidget() );

    notificationWidget_ = new NotificationWidget( scrollArea_ );
    _notificationWidget().move( 10, 10 );
    hSplitter->addWidget( scrollArea_ );
    hSplitter->setCollapsible( 0, false );

    // ref point list widget
    hSplitter->addWidget( listContainer_ = new Widget() );
    listContainer_->setLayout( new QVBoxLayout() );
    listContainer_->layout()->setMargin(0);
    listContainer_->layout()->setSpacing(0);
    listContainer_->layout()->addWidget( refPointListWidget_ = new RefPointListWidget( listContainer_ ) );

    // default width from options, if found
    if( XmlOptions::get().contains( "LIST_WIDTH" ) )
    { listContainer_->setDefaultWidth( XmlOptions::get().get<int>( "LIST_WIDTH" ) ); }

    // event filters
    scrollArea_->installEventFilter( displayWidget_ );
    displayWidget_->installEventFilter( scrollArea_ );

    // assign stretch factors
    hSplitter->setStretchFactor( 0, 1 );
    hSplitter->setStretchFactor( 1, 0 );

    // status bar
    setStatusBar( statusbar_ = new BaseStatusBar( this ) );
    statusbar_->addPermanentWidget( fileEditor_ = new QLabel( statusbar_ ), 1 );
    statusbar_->addLabels( 2, 0 );
    statusbar_->label(0).setAlignment( Qt::AlignCenter );
    statusbar_->label(1).setAlignment( Qt::AlignCenter );
    statusbar_->addClock();

    QLabel& label( statusbar_->label(0) );
    statusbar_->label(0).setMinimumHeight( label.fontMetrics().height() );
    statusbar_->label(1).setMinimumHeight( label.fontMetrics().height() );

    fileEditor_->setTextInteractionFlags( Qt::TextSelectableByMouse|Qt::TextSelectableByKeyboard );

    // splitters
    connect( hSplitter, SIGNAL( splitterMoved( int, int ) ), SLOT( _splitterMoved( void ) ) );
    Application& application( *Singleton::get().application<Application>() );

    // toolbar
    CustomToolBar* toolbar =  new CustomToolBar( tr( "Main" ), this, "TOOLBAR" );
    toolbar->addAction( &application.openAction() );
    toolbar->addAction( &application.saveAction() );
    toolbar->addAction( &application.reloadAction() );
    toolbar->addSeparator();
    toolbar->addAction( undoAction_ );
    toolbar->addAction( redoAction_ );
    toolbar->addSeparator();
    toolbar->addAction( &application.importAction() );
    toolbar->addAction( &application.exportAction() );

    // connections
    connect( scrollArea_->verticalScrollBar(), SIGNAL( valueChanged( int ) ), displayWidget_, SLOT( updatePosition() ) );
    connect( scrollArea_->horizontalScrollBar(), SIGNAL( valueChanged( int ) ), displayWidget_, SLOT( updatePosition() ) );

    // display widget
    connect( displayWidget_, SIGNAL( pointerMoved( QPoint, RefPoint::HoverMode ) ), SLOT( _pointerMoved( QPoint, RefPoint::HoverMode ) ) );
    connect( displayWidget_, SIGNAL( pointAdded( QPointF ) ), SLOT( _addPoint( QPointF ) ) );
    connect( displayWidget_, SIGNAL( pointPasted( QPointF ) ), SLOT( _pastePoint( QPointF ) ) );
    connect( displayWidget_, SIGNAL( pointDragged( QPointF, RefPoint::DragMode ) ), SLOT( _dragPoint( QPointF, RefPoint::DragMode ) ) );
    connect( displayWidget_, SIGNAL( pointSelected( void ) ), SLOT( _selectPoint( void ) ) );

    connect( &displayWidget_->removePointAction(), SIGNAL( triggered( void ) ), SLOT( _removePoint( void ) ) );
    connect( &displayWidget_->editErrorsAction(), SIGNAL( triggered( void ) ), SLOT( _editErrors( void ) ) );
    connect( &displayWidget_->defaultErrorsAction(), SIGNAL( triggered( void ) ), SLOT( _addDefaultErrors( void ) ) );
    connect( &displayWidget_->defaultErrorsSystAction(), SIGNAL( triggered( void ) ), SLOT( _addDefaultErrorsSyst( void ) ) );
    connect( &displayWidget_->copyPointAction(), SIGNAL( triggered( void ) ), SLOT( _copyPoint( void ) ) );

    connect( displayWidget_, SIGNAL( widgetEntered( void ) ), SLOT( _displayWidgetEntered( void ) ) );
    connect( displayWidget_, SIGNAL( widgetLeft( void ) ), SLOT( _displayWidgetLeft( void ) ) );
    connect( displayWidget_, SIGNAL( realZoomFactorChanged( double ) ), &refPointPanel_->zoomWidget(), SLOT( realFactorChanged( double ) ) );

    connect( displayWidget_, SIGNAL( selectionRemoved( const RefPoint::List& ) ), SLOT( _removePoints( const RefPoint::List& ) ) );

    // ref point list
    connect( refPointListWidget_, SIGNAL( modified( void ) ), SLOT( _updatePoints( void ) ) );
    connect( refPointListWidget_, SIGNAL( modified( void ) ), SLOT( _updateSelection( void ) ) );
    connect( refPointListWidget_, SIGNAL( selectionChanged( void ) ), SLOT( _updateSelection( void ) ) );
    connect( refPointListWidget_, SIGNAL( selectionRemoved( const RefPoint::List& ) ), SLOT( _removePoints( const RefPoint::List& ) ) );
    connect( refPointListWidget_, SIGNAL( errorsEdited( RefPoint ) ), SLOT( _editErrors( RefPoint ) ) );
    connect( refPointListWidget_, SIGNAL( defaultErrorsAdded( RefPoint ) ), SLOT( _addDefaultErrors( RefPoint ) ) );
    connect( refPointListWidget_, SIGNAL( defaultErrorsSystAdded( RefPoint ) ), SLOT( _addDefaultErrorsSyst( RefPoint ) ) );

    connect( refPointListWidget_, SIGNAL( clipboardAvailable( bool ) ), &displayWidget_->pastePointAction(), SLOT( setEnabled( bool ) ) );

    // ref point panel
    connect( &refPointPanel_->editActions(), SIGNAL( triggered( QAction* ) ), SLOT( _editRefPoint( QAction* ) ) );
    connect( &refPointPanel_->clearActions(), SIGNAL( triggered( QAction* ) ), SLOT( _clearRefPoint( QAction* ) ) );
    connect( &refPointPanel_->zoomWidget(), SIGNAL( modified( ZoomWidget::Mode, double ) ), displayWidget_, SLOT( zoomModified( ZoomWidget::Mode, double ) ) );
    connect( refPointPanel_, SIGNAL( selectionChanged( void ) ), SLOT( _updateRefPointSelection( void ) ) );
    connect( refPointPanel_, SIGNAL( modified( void ) ), SLOT( _updateRefPoints( void ) ) );
    connect( &application, SIGNAL( configurationChanged() ), SLOT( _updateConfiguration() ) );

    // menu bar
    setMenuBar( menu_ = new Menu( this ) );

    // set focus to display widget
    scrollArea_->setFocus();

    // configuration
    _updateConfiguration();

}

//________________________________________________________________________________________
void MainWindow::setImageRecord( const ImageRecord& record )
{
    Debug::Throw( "MainWindow::setImageRecord.\n" );

    // store
    record_ = record;

    // first clear everything
    clear();
    displayWidget_->setImage( record_.file().file() );

    // update status bar
    fileEditor_->setText( record_.file().file().expand() );

    // restore points
    refPointListWidget_->setCollection( record_.collection() );
    refPointListWidget_->setDataSetIndex( record_.dataSetIndex() );
    refPointListWidget_->updateActions();

    // restore ref points
    const RefPoint::List refPoints( record_.refPoints() );
    for( int i=0; i < qMin( 3, (int)refPoints.size() ); ++i )
    { refPointPanel_->setPoint( i, refPoints[i] ); }

    // flags
    refPointPanel_->setLogX( record_.hasFlag( ImageRecord::LogX ) );
    refPointPanel_->setLogY( record_.hasFlag( ImageRecord::LogY ) );

    // zoom
    displayWidget_->zoomModified( refPointPanel_->zoomWidget().mode(), refPointPanel_->zoomWidget().factor() );

    // update points
    _updatePoints();

    // save in menu
    menu_->recentImagesMenu().setCurrentRecord( record_ );

    // update window title
    QString buffer;
    QTextStream( &buffer ) << record_.file().file().localName() << " - " << record_.file().file().path();
    setWindowTitle( buffer );

}

//________________________________________________________________________________________
bool MainWindow::hasImageRecord( void ) const
{

    Debug::Throw( "MainWindow::hasImageRecord.\n" );
    return !record_.file().file().isEmpty();

}

//________________________________________________________________________________________
ImageRecord MainWindow::imageRecord( void )
{

    Debug::Throw( "MainWindow::imageRecord.\n" );
    record_.setFile( displayWidget_->file() );

    // flags
    record_.setFlag( ImageRecord::LogX, refPointPanel_->hasLogX() );
    record_.setFlag( ImageRecord::LogY, refPointPanel_->hasLogY() );

    // save ref points
    record_.setRefPoints( refPointPanel_->points() );

    // save points
    record_.setCollection( refPointListWidget_->collection() );
    record_.setDataSetIndex( refPointListWidget_->dataSetIndex() );

    return record_;

}

//_______________________________________________________________________________________
void MainWindow::clear( void )
{
    Debug::Throw( "MainWindow::clear.\n" );

    refPointPanel_->clear();
    displayWidget_->clear();
    refPointListWidget_->clear();
    _notificationWidget().clear();
    wasHovered_ = false;
}

//_______________________________________________________________________________________
void MainWindow::updateReferencePoint( int index, const RefPoint& point )
{
    Debug::Throw( "MainWindow::updateReferencePoint.\n" );
    refPointPanel_->setPoint( index, point );
    refPointPanel_->select( index );
    _updatePoints();
    _updateSelection();
}

//_______________________________________________________________________________________
void MainWindow::removeReferencePoint( int index )
{
    Debug::Throw( "MainWindow::removeReferencePoint.\n" );
    refPointPanel_->clear( index );
}

//_______________________________________________________________________________________
void MainWindow::addPoint( int dataSetIndex, const RefPoint& point )
{
    Debug::Throw( "MainWindow::addPoint.\n" );
    refPointListWidget_->setDataSetIndex( dataSetIndex );
    refPointListWidget_->add( point );
    _updatePoints();
    _updateSelection();
}

//_______________________________________________________________________________________
void MainWindow::removePoint( int dataSetIndex, const RefPoint& point )
{
    Debug::Throw( "MainWindow::removePoint.\n" );
    refPointListWidget_->setDataSetIndex( dataSetIndex );
    refPointListWidget_->remove( point );
    _updatePoints();
    _updateSelection();
}

//_______________________________________________________________________________________
void MainWindow::replacePoint( int dataSetIndex, const RefPoint& source, const RefPoint& destination )
{

    Debug::Throw( "MainWindow::replacePoint.\n" );

    refPointListWidget_->setDataSetIndex( dataSetIndex );
    if( refPointListWidget_->replace( source, destination ) )
    {
        _updatePoints();
        _updateSelection();
    }

}

//_______________________________________________________
void MainWindow::timerEvent( QTimerEvent* event )
{

    if( event->timerId() == resizeTimer_.timerId() )
    {

        // stop timer and save time
        resizeTimer_.stop();
        XmlOptions::get().set<int>( "LIST_WIDTH", listContainer_->width() );

    } else return BaseMainWindow::timerEvent( event );

}

//_______________________________________________________________________________________
void MainWindow::_pointerMoved( QPoint position, RefPoint::HoverMode hoverMode )
{
    Debug::Throw( "MainWindow::_pointerMoved.\n" );

    if( !displayWidget_->positionIsValid() ) return;

    // global coordinates
    statusbar_->label(0).setText( RefPoint::PointF( position ).toString( "%.0f" ) );

    // local coordinates
    if( refPointPanel_->isValid() )
    {
        const QPointF global( position/displayWidget_->realZoomFactor() );
        const RefPoint::PointF local( refPointPanel_->toLocal( global ) );
        statusbar_->label(1).setText( local.toString( localFormat_ ) );
    }

    if( hoverMode != RefPoint::NoHover && displayWidget_->dragMode() == RefPoint::NoDrag )
    {

        wasHovered_ = true;
        switch( hoverMode )
        {

            case RefPoint::HoverPoint:
            _notificationWidget().setText( tr( "Press left mouse button to drag the current marker" ) );
            break;

            case RefPoint::HoverRefPoint:
            _notificationWidget().setText( tr( "Press left mouse button to drag the current reference point" ) );
            break;

            default:
            _notificationWidget().setText( tr( "Press left mouse button to drag the error bar of current marker" ) );
            break;

        }

    } else if( wasHovered_ ) {

        wasHovered_ = false;
        _notificationWidget().clear();

    }

}

//_______________________________________________________________________________________
void MainWindow::_displayWidgetEntered( void )
{
    Debug::Throw( "MainWindow::_displayWidgetEntered.\n" );
    if( !refPointPanel_->isValid() ) _notificationWidget().setText( tr( "Press left mouse button to add a new reference point" ) );
    else _notificationWidget().setText( tr( "Press left mouse button to add a new marker" ) );
    wasHovered_ = false;
}

//_______________________________________________________________________________________
void MainWindow::_displayWidgetLeft( void )
{
    Debug::Throw( "MainWindow::_displayWidgetLeft.\n" );
    statusbar_->label(0).clear();
    statusbar_->label(1).clear();
    _notificationWidget().clear();
    wasHovered_ = false;
}

//________________________________________________________________________________________
void MainWindow::_editRefPoint( QAction* action )
{ _editRefPoint( action->actionGroup()->actions().indexOf( action ) ); }

//________________________________________________________________________________________
void MainWindow::_editRefPoint( int index )
{
    Debug::Throw( "MainWindow::_editRefPoint.\n" );

    refPointPanel_->select( index );
    _updateRefPointSelection();

    bool done( false );
    while( !done )
    {
        RefPointDialog dialog( this );
        const RefPoint point( refPointPanel_->point( index ) );
        dialog.setLocalPoint( point.local() );
        if( !dialog.exec() ) return;

        refPointPanel_->setPoint( index, RefPoint( point.global(), dialog.localPoint() ) );

        // check validity
        switch( refPointPanel_->validate() )
        {

            case RefPointPanel::LocalInvalid:
            done = !QuestionDialog( this, tr( "Reference points' local coordinates are aligned.\nUnable to calculate a valid transformation.\nRetry ?" ) ).exec();
            break;

            case RefPointPanel::LogXInvalid:
            case RefPointPanel::LogYInvalid:
            done = !QuestionDialog( this, tr( "Cannot set negative local coordinate in Log scale mode.\nRetry ?" ) ).exec();
            break;

            default:
            done = true;
            break;

        }
    }

}

//________________________________________________________________________________________
void MainWindow::_clearRefPoint( QAction* action )
{ _clearRefPoint( action->actionGroup()->actions().indexOf( action ) ); }

//________________________________________________________________________________________
void MainWindow::_clearRefPoint( int index )
{
    Debug::Throw( "MainWindow::_clearRefPoint.\n" );

    // selecte reference point
    refPointPanel_->select( index );
    _updateRefPointSelection();

    // ask for confirmation
    if( !QuestionDialog( this, tr( "Remove this reference point ?\n Existing markers will be invalidated." ) ).exec() )
    { return; }

    // clear
    refPointPanel_->clear( index );

}

//________________________________________________________________________________________
void MainWindow::_updateRefPoints( void )
{
    Debug::Throw( "MainWindow::_updateRefPoints.\n" );

    const RefPoint::List refPoints( refPointPanel_->points() );
    displayWidget_->setRefPoints( RefPoint::Set::fromList( refPoints ) );

    // get panel validity and update display widget accordingly
    const bool valid( refPointPanel_->isValid() );
    refPointListWidget_->setRefPointsValid( valid );
    displayWidget_->setRefPointsValid( valid );

    // recalculate existing points
    RefPoint::List points( refPointListWidget_->get() );
    for( RefPoint::List::iterator iter = points.begin(); iter != points.end(); ++iter )
    {

        const RefPoint source( *iter );

        // update validity
        iter->setValid( valid );

        // update local position
        if( valid )
        { *iter = refPointPanel_->toLocal( *iter ); }

        // re-insert in model
        refPointListWidget_->replace( source, *iter );

    }

}

//________________________________________________________________________________________
void MainWindow::_updatePoints( void )
{

    Debug::Throw( "MainWindow::_updatePoints.\n" );

    // update collection
    displayWidget_->setCollection( refPointListWidget_->collection() );
    displayWidget_->setDataSetIndex( refPointListWidget_->dataSetIndex() );

    // update current points
    const RefPoint::List& points( refPointListWidget_->get() );
    displayWidget_->setPoints( RefPoint::Set::fromList( points ) );

}

//________________________________________________________________________________________
void MainWindow::_updateSelection( void )
{

    Debug::Throw( "MainWindow::_updateSelection.\n" );
    if( refPointListWidget_->hasSelection() )
    {

        refPointPanel_->clearSelection();
        displayWidget_->setSelection( refPointListWidget_->selection() );

    } else {

        if( refPointPanel_->hasSelection() ) displayWidget_->setSelection( refPointPanel_->selection() );
        else displayWidget_->clearSelection();

    }

}

//________________________________________________________________________________________
void MainWindow::_updateRefPointSelection( void )
{

    Debug::Throw( "MainWindow::_updateRefPointSelection.\n" );
    if( refPointPanel_->hasSelection() )
    {

        refPointListWidget_->clearSelection();
        displayWidget_->setSelection( refPointPanel_->selection() );

    } else {

        if( refPointListWidget_->hasSelection() ) displayWidget_->setSelection( refPointListWidget_->selection() );
        else displayWidget_->clearSelection();

    }

    return;

}

//_______________________________________________________________________________________
void MainWindow::_selectPoint( void )
{

    Debug::Throw() << "MainWindow::_selectPoint - hover mode: " << displayWidget_->hoverMode() << endl;

    if( displayWidget_->hoverMode() == RefPoint::HoverRefPoint )
    {

        refPointListWidget_->clearSelection();
        refPointPanel_->select( displayWidget_->hoverPoint() );

    } else {

        refPointPanel_->clearSelection();

        if( displayWidget_->hoverMode() != RefPoint::NoHover )
        { refPointListWidget_->select( displayWidget_->hoverPoint() ); }

    }

}

//_______________________________________________________________________________________
void MainWindow::_addPoint( QPointF position )
{
    Debug::Throw( "MainWindow::_addPoint.\n" );
    if( refPointPanel_->isValid() )
    {

        const RefPoint point( refPointPanel_->toLocal( RefPoint( position, QPointF() ) ) );

        // add to display widget
        displayWidget_->addPoint( point );

        // clear list selection and select newly added point
        refPointListWidget_->add( point );
        refPointListWidget_->select( point );

        undoStack_->push( new MarkerAddedCommand( this, refPointListWidget_->dataSetIndex(), point ) );

    } else {

        const int index( refPointPanel_->firstInvalidIndex() );

        RefPointDialog dialog( this );
        dialog.setLocalPoint( refPointPanel_->point( index ).local() );

        if( dialog.exec() )
        {

            refPointPanel_->setPoint( index, RefPoint( position, dialog.localPoint() ) );

            // check validity
            switch( refPointPanel_->validate() )
            {

                case RefPointPanel::LocalInvalid:
                InformationDialog( this, tr( "Reference points' local coordinates are aligned.\nUnable to calculate a valid transformation." ) ).exec();
                break;

                case RefPointPanel::GlobalInvalid:
                InformationDialog( this, tr( "Reference points' global coordinates are aligned.\nUnable to calculate a valid transformation." ) ).exec();
                break;

                case RefPointPanel::LogXInvalid:
                case RefPointPanel::LogYInvalid:
                InformationDialog( this, tr( "Cannot set non positive local coordinate in Log scale mode." ) ).exec();
                break;

                default:
                break;

            }

            refPointPanel_->select( index );
            _updateRefPointSelection();

        }

        displayWidget_->updatePosition();
        return;

    }

}

//_______________________________________________________________________________________
void MainWindow::_removePoint( void )
{

    if( displayWidget_->hoverMode() == RefPoint::HoverPoint )
    {

        const RefPoint point( displayWidget_->hoverPoint() );
        refPointListWidget_->remove( point );
        _updatePoints();
        _updateSelection();

        undoStack_->push( new MarkerRemovedCommand( this, refPointListWidget_->dataSetIndex(), point ) );

    } else if( displayWidget_->hoverMode() == RefPoint::HoverRefPoint ) {

        RefPoint point( displayWidget_->hoverPoint() );
        const int index( refPointPanel_->index( point ) );
        if( index >= 0 )
        {
            point = refPointPanel_->point( index );
            refPointPanel_->clear( index );
            undoStack_->push( new RefPointRemovedCommand( this, index, point ) );

        }

    }

}

//_______________________________________________________________________________________
void MainWindow::_removePoints( const RefPoint::List& constPoints )
{
    RefPoint::List points;
    RefPoint::List refPoints;
    foreach( const RefPoint& point, constPoints )
    {
        if( refPointListWidget_->contains( point ) ) points.append( point );
        else if( refPointPanel_->contains( point ) ) refPoints.append( point );
    }

    // add necessary undo commands
    if( !points.empty() )
    {

        // ask for confirmation
        QuestionDialog dialog( this, points.size() > 1 ? tr( "Remove selected markers ?" ):tr( "Remove selected marker ?" ) );
        dialog.setWindowTitle( tr( "Remove Markers - Graphscan" ) );
        if( !dialog.exec() ) return;

        undoStack_->beginMacro( tr( "Remove Markers" ) );
        foreach( const RefPoint& point, points )
        { undoStack_->push( new MarkerRemovedCommand( this, refPointListWidget_->dataSetIndex(), point ) ); }

        undoStack_->endMacro();

    } else if( !refPoints.empty() ) {

        Q_ASSERT( refPoints.size() == 1 );

        // ask for confirmation
        QuestionDialog dialog( this, tr( "Remove selected reference point ?" ) );
        dialog.setWindowTitle( tr( "Remove Reference Point - Graphscan" ) );
        if( !dialog.exec() ) return;

        const int index( refPointPanel_->index( refPoints.front() ) );
        const RefPoint point( refPointPanel_->point( index ) );
        refPointPanel_->clear( index );
        undoStack_->push( new RefPointRemovedCommand( this, index, point ) );

    }

}

//_______________________________________________________________________________________
void MainWindow::_copyPoint( void )
{
    Debug::Throw( "MainWindow::_copyPoint.\n" );
    if( displayWidget_->hoverMode() != RefPoint::HoverPoint ) return;
    refPointListWidget_->copy( displayWidget_->hoverPoint() );
}

//_______________________________________________________________________________________
void MainWindow::_pastePoint( QPointF position )
{
    Debug::Throw( "MainWindow::_pastePoint.\n" );

    RefPoint point( refPointListWidget_->clipboard() );
    if( !point.isValid() ) return;

    point.setGlobal( position );
    point.setLocal( refPointPanel_->toLocal( position ) );
    if( point.hasErrors() || point.hasErrorsSyst() ) point = refPointPanel_->toGlobal( point );

    // execute command
    UndoCommand* command( new MarkerAddedCommand( this, refPointListWidget_->dataSetIndex(), point ) );
    command->setText( tr( "Paste Marker" ) );
    undoStack_->push( command );

    // update selection and display
    refPointListWidget_->select( point );
    _updatePoints();
    _updateSelection();

    return;

}

//_______________________________________________________________________________________
void MainWindow::_editErrors( void )
{

    Debug::Throw( "MainWindow::_editErrors.\n" );
    if( displayWidget_->hoverMode() != RefPoint::HoverPoint ) return;
    return _editErrors( displayWidget_->hoverPoint() );

}

//_______________________________________________________________________________________
void MainWindow::_addDefaultErrors( void )
{
    Debug::Throw( "MainWindow::_addDefaultErrors.\n" );
    if( displayWidget_->hoverMode() != RefPoint::HoverPoint ) return;
    return _addDefaultErrors( displayWidget_->hoverPoint() );
}

//_______________________________________________________________________________________
void MainWindow::_addDefaultErrorsSyst( void )
{
    Debug::Throw( "MainWindow::_addDefaultErrorsSyst.\n" );
    if( displayWidget_->hoverMode() != RefPoint::HoverPoint ) return;
    return _addDefaultErrorsSyst( displayWidget_->hoverPoint() );
}

//_______________________________________________________________________________________
void MainWindow::_editErrors( RefPoint source )
{

    Debug::Throw( "MainWindow::_editErrors.\n" );
    RefPointErrorDialog dialog( this );

    // destination
    RefPoint destination( source );

    // load default errors
    double errorX( XmlOptions::get().get<double>( "DEFAULT_ERROR_X" ) );
    if( errorX > 0 && !destination.hasErrorX() )
    {
        destination.errorX().mode_ = RefPoint::ErrorSymetric;
        destination.errorX().relative_ = true;
        destination.errorX().plus_ = errorX;
    }

    double errorY( XmlOptions::get().get<double>( "DEFAULT_ERROR_X" ) );
    if( errorY > 0 && !destination.hasErrorY() )
    {
        destination.errorY().mode_ = RefPoint::ErrorSymetric;
        destination.errorY().relative_ = true;
        destination.errorY().plus_ = errorY;
    }

    // assign to dialog
    dialog.setPoint( destination );
    if( !dialog.exec() ) return;

    // update point
    destination = dialog.point();
    if( refPointPanel_->isValid() ) destination = refPointPanel_->toGlobal( destination );

    // update display widget
    if( displayWidget_->hoverPoint() == source )
    { displayWidget_->setHoverPoint( destination ); }

    // update model
    UndoCommand* command( new MarkerChangedCommand( this, refPointListWidget_->dataSetIndex(), source, destination ) );
    command->setText( tr( "Edit Marker Errors" ) );
    undoStack_->push( command );

}

//_______________________________________________________________________________________
void MainWindow::_addDefaultErrors( RefPoint source )
{
    Debug::Throw( "MainWindow::_addDefaultErrors.\n" );
    bool errorChanged( false );

    RefPoint destination( source );

    // load default errors
    double errorX( XmlOptions::get().get<double>( "DEFAULT_ERROR_X" ) );
    if( errorX > 0 && !destination.hasErrorX() )
    {
        errorChanged = true;
        destination.errorX().mode_ = RefPoint::ErrorSymetric;
        destination.errorX().relative_ = true;
        destination.errorX().plus_ = errorX;
    }

    double errorY( XmlOptions::get().get<double>( "DEFAULT_ERROR_Y" ) );
    if( errorY > 0 && !destination.hasErrorY() )
    {
        errorChanged = true;
        destination.errorY().mode_ = RefPoint::ErrorSymetric;
        destination.errorY().relative_ = true;
        destination.errorY().plus_ = errorY;
    }

    if( !errorChanged ) return;

    // update global errors
    if( refPointPanel_->isValid() ) destination = refPointPanel_->toGlobal( destination );

    // update display widget
    if( displayWidget_->hoverPoint() == source ) displayWidget_->setHoverPoint( destination );

    // execute command
    UndoCommand* command( new MarkerChangedCommand( this, refPointListWidget_->dataSetIndex(), source, destination ) );
    command->setText( tr( "Edit Marker Errors" ) );
    undoStack_->push( command );

}

//_______________________________________________________________________________________
void MainWindow::_addDefaultErrorsSyst( RefPoint source )
{
    Debug::Throw( "MainWindow::_addDefaultErrorsSyst.\n" );
    bool errorChanged( false );

    RefPoint destination( source );

    // load default errors
    double errorX( XmlOptions::get().get<double>( "DEFAULT_ERROR_SYST_X" ) );
    if( errorX > 0 && !destination.hasErrorSystX() )
    {
        errorChanged = true;
        destination.errorSystX().mode_ = RefPoint::ErrorSymetric;
        destination.errorSystX().relative_ = true;
        destination.errorSystX().plus_ = errorX;
    }

    double errorY( XmlOptions::get().get<double>( "DEFAULT_ERROR_SYST_Y" ) );
    if( errorY > 0 && !destination.hasErrorSystY() )
    {
        errorChanged = true;
        destination.errorSystY().mode_ = RefPoint::ErrorSymetric;
        destination.errorSystY().relative_ = true;
        destination.errorSystY().plus_ = errorY;
    }

    if( !errorChanged ) return;

    // update global errors
    if( refPointPanel_->isValid() ) destination = refPointPanel_->toGlobal( destination );

    // update display widget
    if( displayWidget_->hoverPoint() == source ) displayWidget_->setHoverPoint( destination );

    // execute command
    UndoCommand* command( new MarkerChangedCommand( this, refPointListWidget_->dataSetIndex(), source, destination ) );
    command->setText( tr( "Edit Marker Errors" ) );
    undoStack_->push( command );

}

//_______________________________________________________________________________________
void MainWindow::_addDefaultErrorsToAll( void )
{

    Debug::Throw( "MainWindow::_addDefaultErrorsToAll.\n" );

    const bool valid( refPointPanel_->isValid() );
    undoStack_->beginMacro( tr( "Edit Marker Errors" ) );
    foreach( RefPoint source, refPointListWidget_->get() )
    {

        RefPoint destination( source );

        bool errorChanged( false );

        // load default errors
        double errorX( XmlOptions::get().get<double>( "DEFAULT_ERROR_X" ) );
        if( errorX > 0 && !destination.hasErrorX() )
        {
            errorChanged = true;
            destination.errorX().mode_ = RefPoint::ErrorSymetric;
            destination.errorX().relative_ = true;
            destination.errorX().plus_ = errorX;
        }

        double errorY( XmlOptions::get().get<double>( "DEFAULT_ERROR_Y" ) );
        if( errorY > 0 && !destination.hasErrorY() )
        {
            errorChanged = true;
            destination.errorY().mode_ = RefPoint::ErrorSymetric;
            destination.errorY().relative_ = true;
            destination.errorY().plus_ = errorY;
        }

        if( !errorChanged ) continue;

        // update global errors
        if( valid ) destination = refPointPanel_->toGlobal( destination );

        // execute command
        undoStack_->push( new MarkerChangedCommand( this, refPointListWidget_->dataSetIndex(), source, destination ) );

    }
    undoStack_->endMacro();

    _updatePoints();
    _updateSelection();

}

//_______________________________________________________________________________________
void MainWindow::_addDefaultErrorsSystToAll( void )
{

    Debug::Throw( "MainWindow::_addDefaultErrorsSystToAll.\n" );

    const bool valid( refPointPanel_->isValid() );
    undoStack_->beginMacro( tr( "Edit Marker Errors" ) );
    foreach( RefPoint source, refPointListWidget_->get() )
    {

        RefPoint destination( source );

        bool errorChanged( false );

        // load default errors
        double errorX( XmlOptions::get().get<double>( "DEFAULT_ERROR_SYST_X" ) );
        if( errorX > 0 && !destination.hasErrorSystX() )
        {
            errorChanged = true;
            destination.errorSystX().mode_ = RefPoint::ErrorSymetric;
            destination.errorSystX().relative_ = true;
            destination.errorSystX().plus_ = errorX;
        }

        double errorY( XmlOptions::get().get<double>( "DEFAULT_ERROR_SYST_Y" ) );
        if( errorY > 0 && !destination.hasErrorSystY() )
        {
            errorChanged = true;
            destination.errorSystY().mode_ = RefPoint::ErrorSymetric;
            destination.errorSystY().relative_ = true;
            destination.errorSystY().plus_ = errorY;
        }

        if( !errorChanged ) continue;

        // update global errors
        if( valid ) destination = refPointPanel_->toGlobal( destination );

        // execute command
        undoStack_->push( new MarkerChangedCommand( this, refPointListWidget_->dataSetIndex(), source, destination ) );

    }
    undoStack_->endMacro();

    _updatePoints();
    _updateSelection();

}

//_______________________________________________________________________________________
void MainWindow::_dragPoint( QPointF position, RefPoint::DragMode dragMode )
{

    Debug::Throw() << "MainWindow::_dragPoint - dragMode: " << dragMode << endl;

    // source
    const RefPoint source( displayWidget_->dragPoint() );
    RefPoint destination( source );

    const bool valid( refPointPanel_->isValid() );
    bool errorsChanged( false );
    QString commandLabel;
    switch( dragMode )
    {

        case RefPoint::DragPoint:
        {
            destination.setValid( valid );
            destination.setGlobal( position );

            if( valid )
            {
                // update local point and global errors
                destination.setLocal( refPointPanel_->toLocal( position ) );
                if( destination.hasErrors() || destination.hasErrorsSyst() ) destination = refPointPanel_->toGlobal( destination );
            }

            commandLabel = tr( "Move Marker" );
            break;
        }

        // statistical error x
        case RefPoint::DragErrorXPlus:
        {
            errorsChanged = true;
            destination.errorX().mode_ = RefPoint::ErrorAsymetric;
            destination.errorX().globalPlus_ = position;
            commandLabel = tr( "Edit Marker Errors" );
            break;
        }

        case RefPoint::DragErrorXMinus:
        {
            errorsChanged = true;
            destination.errorX().mode_ = RefPoint::ErrorAsymetric;
            destination.errorX().globalMinus_ = position;
            commandLabel = tr( "Edit Marker Errors" );
            break;
        }

        // statistical error y
        case RefPoint::DragErrorYPlus:
        {
            errorsChanged = true;
            destination.errorY().mode_ = RefPoint::ErrorAsymetric;
            destination.errorY().globalPlus_ = position;
            commandLabel = tr( "Edit Marker Errors" );
            break;
        }

        case RefPoint::DragErrorYMinus:
        {
            errorsChanged = true;
            destination.errorY().mode_ = RefPoint::ErrorAsymetric;
            destination.errorY().globalMinus_ = position;
            commandLabel = tr( "Edit Marker Errors" );
            break;
        }

        // systematic error x
        case RefPoint::DragErrorSystXPlus:
        {
            errorsChanged = true;
            destination.errorSystX().mode_ = RefPoint::ErrorAsymetric;
            destination.errorSystX().globalPlus_ = position;
            commandLabel = tr( "Edit Marker Errors" );
            break;
        }

        case RefPoint::DragErrorSystXMinus:
        {
            errorsChanged = true;
            destination.errorSystX().mode_ = RefPoint::ErrorAsymetric;
            destination.errorSystX().globalMinus_ = position;
            commandLabel = tr( "Edit Marker Errors" );
            break;
        }

        // systematic error y
        case RefPoint::DragErrorSystYPlus:
        {
            errorsChanged = true;
            destination.errorSystY().mode_ = RefPoint::ErrorAsymetric;
            destination.errorSystY().globalPlus_ = position;
            commandLabel = tr( "Edit Marker Errors" );
            break;
        }

        case RefPoint::DragErrorSystYMinus:
        {
            errorsChanged = true;
            destination.errorSystY().mode_ = RefPoint::ErrorAsymetric;
            destination.errorSystY().globalMinus_ = position;
            commandLabel = tr( "Edit Marker Errors" );
            break;
        }

        case RefPoint::DragRefPoint:
        {
            if( !refPointPanel_->hasSelection() ) return;

            const int index( refPointPanel_->selectedIndex() );
            const RefPoint source( refPointPanel_->point( index ) );
            destination.setGlobal( position );
            destination.setLocal( source.local() );

            // execute command
            UndoCommand* command( new RefPointChangedCommand( this, index, source, destination ) );
            command->setText( tr( "Move Reference Point" ) );
            undoStack_->push( command );

            break;
        }

        default: break;

    }

    if( dragMode != RefPoint::DragRefPoint )
    {

        // update errors
        if( errorsChanged && valid )
        {
            // update global to local and back, to have correct global error positions
            destination = refPointPanel_->toLocal( destination );
            destination = refPointPanel_->toGlobal( destination );
        }

        // execute command
        UndoCommand* command( new MarkerChangedCommand( this, refPointListWidget_->dataSetIndex(), source, destination ) );
        if( !commandLabel.isEmpty() ) command->setText( commandLabel );
        undoStack_->push( command );

    }

    // update display
    displayWidget_->setDragPoint( destination );
    displayWidget_->setDragMode( dragMode );

}

//_______________________________________________________________________________________
void MainWindow::_installActions( void )
{

    addAction( defaultErrorsAction_ = new QAction( IconEngine::get( ICONS::EDIT ), tr( "Add Default Statistical Errors" ), this ) );
    connect( defaultErrorsAction_, SIGNAL( triggered( void ) ), SLOT( _addDefaultErrorsToAll( void ) ) );

    addAction( defaultErrorsSystAction_ = new QAction( IconEngine::get( ICONS::EDIT ), tr( "Add Default Systematic Errors" ), this ) );
    connect( defaultErrorsSystAction_, SIGNAL( triggered( void ) ), SLOT( _addDefaultErrorsSystToAll( void ) ) );

    addAction( undoAction_ = undoStack_->createUndoAction( this ) );
    undoAction_->setIcon( IconEngine::get( ICONS::UNDO ) );
    undoAction_->setShortcut( QKeySequence::Undo );

    addAction( redoAction_ = undoStack_->createRedoAction( this ) );
    redoAction_->setIcon( IconEngine::get( ICONS::REDO ) );
    redoAction_->setShortcut( QKeySequence::Redo );

    new QShortcut( Qt::Key_Up, this, SLOT( _movePointsUp( void ) ) );
    new QShortcut( Qt::Key_Down, this, SLOT( _movePointsDown( void ) ) );
    new QShortcut( Qt::Key_Left, this, SLOT( _movePointsLeft( void ) ) );
    new QShortcut( Qt::Key_Right, this, SLOT( _movePointsRight( void ) ) );
    return;

}

//_______________________________________________________________________________________
void MainWindow::_movePoints( Qt::Key key )
{

    // calculate offset based on the key
    RefPoint::PointF offset;

    switch( key )
    {
        case Qt::Key_Up: offset = RefPoint::PointF( 0, -1.0/displayWidget_->realZoomFactor() ); break;
        case Qt::Key_Down: offset = RefPoint::PointF( 0, 1.0/displayWidget_->realZoomFactor() ); break;
        case Qt::Key_Left: offset = RefPoint::PointF( -1.0/displayWidget_->realZoomFactor(), 0 ); break;
        case Qt::Key_Right: offset = RefPoint::PointF( 1.0/displayWidget_->realZoomFactor(), 0 ); break;
        default: return;
    }

    // find selected points
    RefPoint::List selection( refPointListWidget_->selection() );
    if( !selection.empty() )
    {

        if( selection.size() > 1 ) undoStack_->beginMacro( tr( "Move Markers" ) );

        const bool valid( refPointPanel_->isValid() );
        foreach( const RefPoint& source, selection )
        {
            RefPoint destination( source );
            destination.setValid( valid );
            destination.setGlobal( destination.global() + offset );
            if( valid )
            {
                destination.setLocal( refPointPanel_->toLocal( destination.global() ) );
                if( destination.hasErrors() || destination.hasErrorsSyst() ) destination = refPointPanel_->toGlobal( destination );
            }

            // execute command
            undoStack_->push( new MarkerChangedCommand( this, refPointListWidget_->dataSetIndex(), source, destination ) );

        }

        if( selection.size() > 1 ) undoStack_->endMacro();

    } else if( refPointPanel_->hasSelection() ) {

        const int index( refPointPanel_->selectedIndex() );
        const RefPoint source( refPointPanel_->point( index ) );
        RefPoint destination( source );
        destination.setGlobal( destination.global() + offset );
        destination.setLocal( destination.local() );

        // execute command
        undoStack_->push( new RefPointChangedCommand( this, index, source, destination ) );

    }

    _updatePoints();
    _updateSelection();

    return;

}

//________________________________________________________
void MainWindow::_splitterMoved( void )
{
    Debug::Throw( "MainWindow::_splitterMoved.\n" );
    resizeTimer_.start( 200, this );
}

//_______________________________________________________________________________________
void MainWindow::_updateConfiguration( void )
{
    Debug::Throw( "MainWindow::_updateConfiguration.\n" );

    // change scrollarea palette
    // this does not work with oxygen because overridden.
    QPalette palette( scrollArea_->palette() );
    palette.setColor( QPalette::Window, QColor( XmlOptions::get().get<BASE::Color>( "SCROLL_AREA_COLOR" ) ) );
    scrollArea_->ensurePolished();
    scrollArea_->viewport()->setPalette( palette );
    scrollArea_->viewport()->setBackgroundRole( QPalette::Window );
    scrollArea_->viewport()->setAutoFillBackground( true );

    defaultErrorsAction().setEnabled( XmlOptions::get().get<double>( "DEFAULT_ERROR_X" ) > 0 || XmlOptions::get().get<double>( "DEFAULT_ERROR_Y" ) > 0 );

    // get formats
    localFormat_ = XmlOptions::get().get<QString>( "LOCAL_FORMAT" );

}

//_______________________________________________________________________________________
void MainWindow::Widget::setDefaultWidth( int value )
{ defaultWidth_ = value; }

//_______________________________________________________________________________________
void MainWindow::Widget::setDefaultHeight( int value )
{ defaultHeight_ = value; }

//____________________________________________
QSize MainWindow::Widget::sizeHint( void ) const
{ return (defaultWidth_>0 || defaultHeight_>0 ) ? QSize( defaultWidth_, defaultHeight_ ):QWidget::sizeHint(); }
