// $Id: RecentImagesMenu.cpp,v 1.9 2013/03/12 08:47:46 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "RecentImagesMenu.h"

#include "BaseIcons.h"
#include "FileRecordBaseProperties.h"
#include "IconEngine.h"
#include "QuestionDialog.h"
#include "XmlOptions.h"

#include <algorithm>

//_______________________________________________
RecentImagesMenu::RecentImagesMenu( QWidget *parent, ImageRecordList& files ):
    QMenu( parent ),
    Counter( "RecentImagesMenu" ),
    imageList_( &files )
{
    Debug::Throw( "RecentImagesMenu::RecentImagesMenu.\n" );

    setTitle( tr( "Open Recent" ) );
    connect( this, SIGNAL( triggered( QAction* ) ), SLOT( _open( QAction* ) ) );
    connect( this, SIGNAL( aboutToShow( void ) ), imageList_, SLOT( checkValidFiles( void ) ) );
    connect( this, SIGNAL( aboutToShow( void ) ), SLOT( _loadFiles( void ) ) );
    connect( imageList_, SIGNAL( validFilesChecked( void ) ), SLOT( _updateActions( void ) ) );

    // icons
    setIcon( IconEngine::get( ICONS::OPEN ) );

    actionGroup_ = new QActionGroup( this );
    actionGroup_->setExclusive( true );

    addAction( cleanAction_ = new QAction( IconEngine::get( ICONS::DELETE ), tr( "Clean" ), this ) );
    connect( cleanAction_, SIGNAL( triggered() ), SLOT( _clean() ) );
    cleanAction_->setEnabled( false );
    addSeparator();

}

//______________________________________
RecentImagesMenu::~RecentImagesMenu( void )
{ Debug::Throw( "RecentImagesMenu::~RecentImagesMenu.\n" ); }

//______________________________________
bool RecentImagesMenu::openLastValidFile( void )
{

    Debug::Throw( "RecentImagesMenu::openLastValidFile.\n" );
    ImageRecord record( imageList_->lastValidFile() );
    if( record.file().file().isEmpty() ) return false;

    emit fileSelected( record );
    return true;

}

//______________________________________
void RecentImagesMenu::setCurrentRecord( const ImageRecord& record )
{ currentRecord_ = ImageRecord( record.file().file(), record.version() ); }

//______________________________________
void RecentImagesMenu::_updateActions( void )
{

    Debug::Throw( "RecentImagesMenu::_updateActions.\n" );

    // set actions enability
    ImageRecord::List records( imageList_->records() );
    for( ActionMap::iterator iter = actions_.begin(); iter != actions_.end(); ++iter )
    {

        ImageRecord::List::const_iterator found = std::find_if(
            records.begin(),
            records.end(),
            ImageRecord::SameFileFTor( iter.value() ) );
        if( found == records.end() )
        { continue; }

        iter.value().file().setValid( found->file().isValid() );
        iter.key()->setEnabled( found->file().isValid() );

    }

    cleanAction_->setEnabled( imageList_->cleanEnabled() );

}

//______________________________________
void RecentImagesMenu::_clean( void )
{
    if( !imageList_->check() && !QuestionDialog( this, tr( "clear list ?" ) ).exec() ) return;
    else if( imageList_->check() && !QuestionDialog( this, tr( "Remove invalid or duplicated files from list ?" ) ).exec() ) return;
    imageList_->clean();
}

//_______________________________________________
void RecentImagesMenu::_open( QAction* action )
{

    Debug::Throw( "RecentImagesMenu::_Open.\n" );

    // find Action in map
    ActionMap::iterator iter( actions_.find( action ) );
    if( iter == actions_.end() ) return;
    emit fileSelected( iter.value() );

}

//_______________________________________________
void RecentImagesMenu::_loadFiles( void )
{
    Debug::Throw( "RecentImagesMenu::_loadFiles.\n" );

    // run thread to check file validity
    cleanAction_->setEnabled( imageList_->cleanEnabled() );

    // clear menu an actions map
    for( ActionMap::iterator iter = actions_.begin(); iter != actions_.end(); ++iter )
    { delete iter.key(); }
    actions_.clear();

    // redo all actions
    ImageRecord::List records( imageList_->records() );
    if( XmlOptions::get().get<bool>("SORT_FILES_BY_DATE") ) { std::sort( records.begin(), records.end(), ImageRecord::FirstOpenFTor() ); }
    else { std::sort( records.begin(), records.end(), ImageRecord::FileFTor() ); }

    // retrieve stored file record
    foreach( const ImageRecord& record, records )
    {

        QString label( record.file().file() );
        if( record.version() > 0 ) label += QString( " (%1)" ).arg( record.version()+1 );
        QAction* action = addAction( label );

        // add icon
        if( record.file().hasProperty( FileRecordProperties::ICON ) ) { action->setIcon( IconEngine::get( record.file().property( FileRecordProperties::ICON ) ) ); }

        // check action if match file
        action->setCheckable( true );
        action->setChecked( record == currentRecord() );
        actionGroup_->addAction( action );

        if( imageList_->check() ) action->setEnabled( record.file().file().size() && record.file().isValid() );
        actions_.insert( action, ImageRecord( record.file().file(), record.version() ) );
    }

}
