// $Id: RefPoint.cpp,v 1.13 2013/02/14 19:58:31 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "RefPoint.h"
#include "Debug.h"
#include "XmlDef.h"

//__________________________________________________________________
RefPoint::PointF::PointF( const QDomElement& element ):
    Counter( "RefPoint::PointF" )
{

    Debug::Throw( "RefPoint::PointF::PointF (dom)\n" );

    QDomNamedNodeMap attributes( element.attributes() );
    for( unsigned int i=0; i<attributes.length(); i++ )
    {
        QDomAttr attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;

        QString name( attribute.name() );
        QString value( attribute.value() );
        if( name == XML::X ) setX( value.toDouble() );
        else if( name == XML::Y ) setY( value.toDouble() );
        else Debug::Throw(0) << "RefPoint::PointF::PointF - unrecognized entry attribute: \"" << name << "\"" << endl;

    }

}

//__________________________________________________________________
QDomElement RefPoint::PointF::domElement( const QString& name, QDomDocument& parent ) const
{

    Debug::Throw() << "RefPoint::PointF::domElement - name: " << name << endl;

    QDomElement out = parent.createElement( name );
    out.setAttribute( XML::X, x() );
    out.setAttribute( XML::Y, y() );
    return out;
}

//__________________________________________________________________
QString RefPoint::PointF::toString( const QString& format ) const
{
    if( format.isEmpty() )
    {
        QString buffer;
        QTextStream( &buffer ) << "(" << x() << "," << y() << ")";
        return buffer;
    } else {
        QString buffer;
        QTextStream( &buffer )
            << "("
            << QString().sprintf( format.toLatin1().constData(), x() ) << ","
            << QString().sprintf( format.toLatin1().constData(), y() ) << ")";
        return buffer;
    }

}

//__________________________________________________________________
RefPoint::Error::Error( const QDomElement& element ):
    Counter( "RefPoint::Error" ),
    mode_( NoError ),
    plus_( 0 ),
    minus_( 0 ),
    relative_( false )
{

    Debug::Throw( "RefPoint::Error::Error (dom).\n" );

    QDomNamedNodeMap attributes( element.attributes() );
    for( unsigned int i=0; i<attributes.length(); i++ )
    {
        QDomAttr attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;

        QString name( attribute.name() );
        QString value( attribute.value() );
        if( name == XML::ERROR_MODE ) mode_ = (ErrorMode) value.toUInt();
        else if( name == XML::ERROR_PLUS ) plus_ = value.toDouble();
        else if( name == XML::ERROR_MINUS ) minus_ = value.toDouble();
        else if( name == XML::ERROR_RELATIVE ) relative_ = value.toUInt();
        else Debug::Throw(0) << "RefPoint::Error::Error - unrecognized attribute: \"" << name << "\"" << endl;
    }

    // parse children elements
    for(QDomNode childNode = element.firstChild(); !childNode.isNull(); childNode = childNode.nextSibling() )
    {
        QDomElement childElement = childNode.toElement();
        if( childElement.isNull() ) continue;
        QString tagName( childElement.tagName() );

        if( tagName == XML::ERROR_GLOBAL_PLUS ) globalPlus_ = RefPoint::PointF( childElement );
        else if( tagName == XML::ERROR_GLOBAL_MINUS ) globalMinus_ = RefPoint::PointF( childElement );
        else Debug::Throw(0) << "RefPoint::Error::Error - unrecognized child " << childElement.tagName() << endl;
    }

}

//__________________________________________________________________
QDomElement RefPoint::Error::domElement( const QString& name, QDomDocument& parent ) const
{

    Debug::Throw() << "RefPoint::Error::domElement - name: " << name << endl;

    QDomElement out = parent.createElement( name );
    out.setAttribute( XML::ERROR_MODE, mode_ );
    out.setAttribute( XML::ERROR_PLUS, plus_ );
    out.setAttribute( XML::ERROR_MINUS, minus_ );
    out.setAttribute( XML::ERROR_RELATIVE, relative_ );

    out.appendChild( globalPlus_.domElement( XML::ERROR_GLOBAL_PLUS, parent ) );
    out.appendChild( globalMinus_.domElement( XML::ERROR_GLOBAL_MINUS, parent ) );

    return out;
}

//__________________________________________________________________
RefPoint::List::List( const QDomElement& element ):
    Counter( "RefPoint::List" )
{

    Debug::Throw( "RefPoint::List::List (dom).\n" );

    // parse children elements
    for(QDomNode childNode = element.firstChild(); !childNode.isNull(); childNode = childNode.nextSibling() )
    {

        QDomElement childElement = childNode.toElement();
        if( childElement.isNull() ) continue;
        QString tagName( childElement.tagName() );

        if( tagName == XML::POINT ) append( RefPoint( childElement ) );
        else Debug::Throw(0) << "RefPoint::List::List - unrecognized child " << childElement.tagName() << endl;

    }

}

//__________________________________________________________________
QDomElement RefPoint::List::domElement( const QString& name, QDomDocument& parent ) const
{

    Debug::Throw() << "RefPoint::List::domElement - name: " << name << endl;

    QDomElement out( parent.createElement( name ) );
    for( const_iterator iter = begin(); iter != end(); ++iter )
    { out.appendChild( iter->domElement( XML::POINT, parent ) ); }

    return out;
}

//__________________________________________________________________
RefPoint::RefPoint( const QDomElement& element ):
    Counter( "RefPoint" ),
    valid_( true )
{
    Debug::Throw( "RefPoint::RefPoint (dom).\n" );

    // parse attributes
    QDomNamedNodeMap attributes( element.attributes() );
    for( unsigned int i=0; i<attributes.length(); i++ )
    {
        QDomAttr attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;

        QString name( attribute.name() );
        QString value( attribute.value() );
        if( name == XML::VALID ) setValid( value.toUInt() );
        else Debug::Throw(0) << "RefPoint::RefPoint - unrecognized attribute: \"" << name << "\"" << endl;

    }

    // parse children elements
    for(QDomNode childNode = element.firstChild(); !childNode.isNull(); childNode = childNode.nextSibling() )
    {
        QDomElement childElement = childNode.toElement();
        if( childElement.isNull() ) continue;
        QString tagName( childElement.tagName() );

        if( tagName == XML::GLOBAL ) setGlobal( PointF( childElement ) );
        else if( tagName == XML::LOCAL ) setLocal( PointF( childElement ) );
        else if( tagName == XML::ERROR_X ) errorX_ = Error( childElement );
        else if( tagName == XML::ERROR_Y ) errorY_ = Error( childElement );
        else if( tagName == XML::ERROR_SYST_X ) errorSystX_ = Error( childElement );
        else if( tagName == XML::ERROR_SYST_Y ) errorSystY_ = Error( childElement );
        else Debug::Throw(0) << "RefPoint::RefPoint - unrecognized child " << childElement.tagName() << endl;
    }

}

//__________________________________________________________________
QDomElement RefPoint::domElement( const QString& name, QDomDocument& parent ) const
{

    Debug::Throw() << "RefPoint::domElement - name: " << name << endl;

    QDomElement out( parent.createElement( name ) );
    out.setAttribute( XML::VALID, QString::number( isValid() ) );

    // local position
    out.appendChild( local_.domElement( XML::LOCAL, parent ) );

    // global position
    out.appendChild( global_.domElement( XML::GLOBAL, parent ) );

    // statistical
    if( hasErrorX() ) out.appendChild( errorX_.domElement( XML::ERROR_X, parent ) );
    if( hasErrorY() ) out.appendChild( errorY_.domElement( XML::ERROR_Y, parent ) );

    // systematic
    if( hasErrorSystX() ) out.appendChild( errorSystX_.domElement( XML::ERROR_SYST_X, parent ) );
    if( hasErrorSystY() ) out.appendChild( errorSystY_.domElement( XML::ERROR_SYST_Y, parent ) );

    return out;

}
