#ifndef RefPoint_h
#define RefPoint_h

// $Id: RefPoint.h,v 1.18 2013/02/14 09:02:46 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "Counter.h"
#include "QOrderedSet.h"

#include <QDomDocument>
#include <QDomElement>
#include <QList>
#include <QPair>
#include <QPointF>
#include <QTextStream>

#include <cmath>

class RefPoint: public Counter
{

    public:

    //! error mode
    enum ErrorMode
    {
        NoError,
        ErrorSymetric,
        ErrorAsymetric
    };

    //! hover mode
    enum HoverMode
    {

        NoHover,

        // points
        HoverPoint,
        HoverRefPoint,

        // statistical errors
        HoverErrorXPlus,
        HoverErrorXMinus,
        HoverErrorYPlus,
        HoverErrorYMinus,

        // systematic errors
        HoverErrorSystXPlus,
        HoverErrorSystXMinus,
        HoverErrorSystYPlus,
        HoverErrorSystYMinus

    };

    //! drag mode
    enum DragMode
    {

        NoDrag,

        // points
        DragPoint,
        DragRefPoint,

        // statistical errors
        DragErrorXPlus,
        DragErrorXMinus,
        DragErrorYPlus,
        DragErrorYMinus,

        // systematic errors
        DragErrorSystXPlus,
        DragErrorSystXMinus,
        DragErrorSystYPlus,
        DragErrorSystYMinus

    };

    //! implements less than operators on QPointF
    class PointF: public QPointF, public Counter
    {

        public:

        typedef QSet<PointF> Set;

        //! constructor
        PointF( void ):
            Counter( "RefPoint::PointF" )
        {}

        //! constructor
        PointF( double x, double y ):
            QPointF( x, y ),
            Counter( "RefPoint::PointF" )
        {}


        //! constructor
        PointF( const QPointF& other ):
            QPointF( other ),
            Counter( "RefPoint::PointF" )
            {}

        //! creator from DOM
        PointF( const QDomElement& );

        //! get DOM element
        QDomElement domElement( const QString&, QDomDocument& ) const;

        //! less than operator
        bool operator < (const QPointF& other ) const
        {
            if( x() != other.x() ) return x() < other.x();
            else return y() < other.y();
        }

        //! length
        double length( void ) const
        { return std::sqrt( x()*x() + y()*y() ); }

        //! distance
        double length( const QPointF& other ) const
        { return std::sqrt( (x() - other.x())*(x() - other.x()) + (y() - other.y())*(y() - other.y()) ); }

        //! normalize point
        PointF normalized( void ) const
        { return PointF( (*this)*(1.0/length()) ); }

        //! rotate point
        /*!
        returns rotated point such that x is along argument, and y perpendicular to it.
        assumes that direction.length() is unity
        */
        PointF rotated( QPointF direction ) const
        { return PointF( direction.x()*x() + direction.y()*y(), -direction.y()*x() + direction.x()*y() ); }

        //! conversion to string
        QString toString( const QString& = QString() ) const;

    };

    //! errors on point
    class Error: public Counter
    {

        public:

        //! constructor
        Error( void ):
            Counter( "RefPoint::Error" ),
            mode_( NoError ),
            plus_( 0 ),
            minus_( 0 ),
            relative_( false )
        {}

        //! creator from DOM
        Error( const QDomElement& );

        //! get DOM element
        QDomElement domElement( const QString&, QDomDocument& ) const;

        //! clear
        void clear( void )
        { mode_ = NoError; }

        //! validity
        bool isValid( void ) const
        {
            return
                mode_ != NoError &&
                (plus_ != 0 || minus_ != 0) &&
                (globalPlus_ != globalMinus_ );
        }

        ErrorMode mode_;
        double plus_;
        double minus_;
        bool relative_;

        RefPoint::PointF globalPlus_;
        RefPoint::PointF globalMinus_;

    };

    //! list of reference points
    class List: public QList<RefPoint>, public Counter
    {
        public:

        //! default constructor
        List( void ):
            Counter( "RefPoint::List" )
        {}

        //! copy constructor
        List( const QList<RefPoint>& other ):
            QList<RefPoint>( other ),
            Counter( "RefPoint::List" )
            {}

        //! creator from DOM
        List( const QDomElement& );

        //! get DOM element
        QDomElement domElement( const QString&, QDomDocument& ) const;

    };


    typedef QListIterator<RefPoint> ListIterator;

    //! shortcut to list of RefPoint
    typedef QOrderedSet<RefPoint> Set;

    //! shortcut to dataset
    typedef List DataSet;

    //! shortcut to list of lists
    typedef QList<RefPoint::List> Collection;

    //! pair
    typedef QPair<RefPoint, RefPoint> Pair;
    typedef QList<Pair> PairList;

    //! constructor
    RefPoint( void ):
        Counter( "RefPoint" ),
        valid_( true )
    {}

    //! constructor
    RefPoint( QPointF global, QPointF local ):
        Counter( "RefPoint" ),
        valid_( true ),
        global_( global ),
        local_( local )
    {}

    //! destructor
    virtual ~RefPoint( void )
    {}

    //! creator from DOM
    RefPoint( const QDomElement& );

    //! get DOM element
    QDomElement domElement( const QString&, QDomDocument& ) const;

    //! equal to operator
    bool operator == (const RefPoint& other ) const
    { return other.global_ == global_; }

    //! less than operatir
    bool operator < (const RefPoint& other ) const
    { return PointF( global_ ) < other.global_; }

    //!@name accessors
    //@{

    //! global point
    const PointF& global( void ) const
    { return global_; }

    //! local point
    const PointF& local( void ) const
    { return local_; }

    //! validity
    bool isValid( void ) const
    { return valid_; }

    //! errors
    bool hasErrorX( void ) const
    { return errorX_.isValid(); }

    //! errors
    bool hasErrorY( void ) const
    { return errorY_.isValid(); }

    //! errors
    bool hasErrors( void ) const
    { return hasErrorX() || hasErrorY(); }

    //! horizontal error
    const Error& errorX( void ) const
    { return errorX_; }

    //! vertical error
    const Error& errorY( void ) const
    { return errorY_; }

    //! errors
    bool hasErrorSystX( void ) const
    { return errorSystX_.isValid(); }

    //! errors
    bool hasErrorSystY( void ) const
    { return errorSystY_.isValid(); }

    //! errors
    bool hasErrorsSyst( void ) const
    { return hasErrorSystX() || hasErrorSystY(); }

    //! horizontal error
    const Error& errorSystX( void ) const
    { return errorSystX_; }

    //! vertical error
    const Error& errorSystY( void ) const
    { return errorSystY_; }

    //@}

    //!@name modifiers
    //@{

    //! global point
    bool setGlobal( const QPointF& position )
    {
        if( global_ == position ) return false;
        global_ = position;
        return true;
    }

    //! local point
    bool setLocal( const QPointF& position )
    {
        if( local_ == position ) return false;
        local_ = position;
        return true;
    }

    //! horizontal error
    Error& errorX( void )
    { return errorX_; }

    //! vertical error
    Error& errorY( void )
    { return errorY_; }

    //! horizontal systematic error
    Error& errorSystX( void )
    { return errorSystX_; }

    //! vertical systematic error
    Error& errorSystY( void )
    { return errorSystY_; }

    //! validity
    void setValid( bool value )
    { valid_ = value; }

    //! clear
    void clear( void )
    {
        global_ = PointF();
        local_ = PointF();

        // errors
        errorX_.clear();
        errorY_.clear();

        // systematc errors
        errorSystX_.clear();
        errorSystY_.clear();
        valid_ = false;
    }

    //@}

    //! find refpoint matching local position
    class SameLocalFTor
    {
        public:

        //! constructor
        SameLocalFTor( QPointF local ):
            local_( local )
            {}

        //! prediction
        bool operator () (const RefPoint& other ) const
        { return other.local() == local_; }

        private:

        //! predicate
        QPointF local_;

    };


    //! find refpoint matching global position
    class SameGlobalFTor
    {
        public:

        //! constructor
        SameGlobalFTor( QPointF global ):
            global_( global )
            {}

        //! prediction
        bool operator () (const RefPoint& other ) const
        { return other.global() == global_; }

        private:

        //! predicate
        QPointF global_;

    };

    private:

    //! validity flag
    bool valid_;

    //! global point
    PointF global_;

    //! local point
    PointF local_;

    //!@name error
    //@{
    Error errorX_;
    Error errorY_;
    Error errorSystX_;
    Error errorSystY_;
    //@}

};

#endif
