// $Id: RefPointListWidget.cpp,v 1.30 2013/03/22 09:41:40 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "RefPointListWidget.h"

#include "BaseContextMenu.h"
#include "CustomDialog.h"
#include "CustomToolBar.h"
#include "Icons.h"
#include "IconSize.h"
#include "IconEngine.h"
#include "PrintListDialog.h"
#include "QuestionDialog.h"
#include "RefPointFormatDialog.h"
#include "Singleton.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QHeaderView>
#include <QLayout>

//____________________________________________________________
RefPointListWidget::RefPointListWidget( QWidget* parent ):
    QWidget( parent ),
    dataSetIndex_( 0 )
{

    QVBoxLayout* vLayout( new QVBoxLayout() );
    vLayout->setMargin(0);
    vLayout->setSpacing(2);
    setLayout( vLayout );

    vLayout->addWidget( dataSetWidget_ = new DataSetWidget( this ) );
    connect( dataSetWidget_, SIGNAL( dataSetAdded( void ) ), SLOT( _addDataSet( void ) ) );
    connect( dataSetWidget_, SIGNAL( dataSetRemoved( void ) ), SLOT( _removeDataSet( void ) ) );
    connect( dataSetWidget_, SIGNAL( dataSetSelected( int ) ), SLOT( _selectDataSet( int ) ) );

    QHBoxLayout* hLayout( new QHBoxLayout() );
    hLayout->setSpacing(0);
    hLayout->setMargin(0);
    vLayout->addLayout( hLayout );

    hLayout->addWidget( list_ = new TreeView( this ), 1 );
    list_->setContextMenuPolicy( Qt::CustomContextMenu );
    list_->setOptionName( "REFPOINT_LIST" );
    list_->setModel( &model_ );
    list_->setItemMargin( 2 );
    list_->setSelectionMode( QAbstractItemView::ContiguousSelection );
    list_->setSortingEnabled( false );

    connect( list_->selectionModel(), SIGNAL( selectionChanged( const QItemSelection& , const QItemSelection& ) ), SLOT( updateActions() ) );
    connect( list_->selectionModel(), SIGNAL( selectionChanged( const QItemSelection& , const QItemSelection& ) ), SIGNAL( selectionChanged() ) );
    connect( list_, SIGNAL( customContextMenuRequested( const QPoint& ) ), SLOT( _updateContextMenu( const QPoint& ) ) );

    connect( &model_, SIGNAL( rowsInserted( const QModelIndex&, int, int ) ), SLOT(updateActions() ) );
    connect( &model_, SIGNAL( rowsRemoved( const QModelIndex&, int, int ) ), SLOT(updateActions() ) );

    // actions
    _installActions();

    // toolbar
    CustomToolBar* toolbar = new CustomToolBar( tr( "Reference Points" ), this, "REFPOINT_TOOLBAR" );
    toolbar->setMovable( false );
    toolbar->setTransparent( true );
    toolbar->setOrientation( Qt::Vertical );
    toolbar->setIconSize( IconSize( IconSize::Small ) );

    toolbar->addAction( removeAction_ );
    toolbar->addAction( moveUpAction_ );
    toolbar->addAction( moveDownAction_ );
    toolbar->addAction( sortAction_ );
    toolbar->addAction( printAction_ );
    toolbar->addAction( refPointFormatAction_ );

    hLayout->addWidget( toolbar, 0 );

    updateActions();

    connect( Singleton::get().application(), SIGNAL( configurationChanged() ), SLOT( _updateConfiguration() ) );
    connect( this, SIGNAL( refPointFormatChanged( void ) ), SLOT( _updateConfiguration( void ) ) );

    _updateConfiguration();

}

//____________________________________________________________
RefPoint::List RefPointListWidget::get( void ) const
{ return model_.get(); }


//____________________________________________________________
RefPoint::Collection RefPointListWidget::collection( void ) const
{

    RefPoint::Collection out( dataSets_ );
    if( dataSetIndex_ < (int) out.size() )
    { out[dataSetIndex_] = model_.get(); }

    return out;
}

//____________________________________________________________
bool RefPointListWidget::hasSelection( void ) const
{ return !list_->selectionModel()->selectedRows().isEmpty(); }

//____________________________________________________________
RefPoint::List RefPointListWidget::selection( void ) const
{
    QModelIndexList selectedIndexes( list_->selectionModel()->selectedRows() );
    return model_.get( selectedIndexes );
}

//____________________________________________________________
void RefPointListWidget::setCollection( const RefPoint::Collection& collection )
{

    Debug::Throw( "RefPointListWidget::setCollection.\n" );

    dataSets_ = collection;
    if( dataSetIndex_ < (int) dataSets_.size() )
    {

        model_.set( dataSets_[dataSetIndex_] );
        list_->resizeColumns();

    } else model_.clear();

    // clear selection
    list_->clearSelection();

    dataSetWidget_->setCollection( collection );

}

//____________________________________________________________
void RefPointListWidget::addDataSet( const RefPoint::List& points )
{

    Debug::Throw( "RefPointListWidget::addDataSet.\n" );

    // store in collection
    dataSets_ << points;

    // pass to model if index matches
    if( (int) dataSets_.size() == dataSetIndex_ + 1 )
    {
        model_.set( points );
        list_->resizeColumns();
    }

}

//____________________________________________________________
void RefPointListWidget::setDataSetIndex( int value )
{

    Debug::Throw() << "RefPointListWidget::setDataSetIndex - " << value << endl;

    // check if changed
    if( value == dataSetIndex_ ) return;

    // first store current dataset if valid
    if( dataSetIndex_ < (int) dataSets_.size() )
    { dataSets_[dataSetIndex_] = model_.get(); }

    // update index
    dataSetIndex_ = value;

    // update model
    model_.clear();
    if( dataSetIndex_ < (int) dataSets_.size() )
    {
        model_.set( dataSets_[dataSetIndex_] );
        list_->resizeColumns();
    }

    // clear selection
    list_->clearSelection();

    // update dataset widget
    dataSetWidget_->select( dataSetIndex_ );

}

//____________________________________________________________
void RefPointListWidget::clear( void )
{
    Debug::Throw( "RefPointListWidget::clear.\n" );
    dataSets_.clear();
    model_.clear();
    clipboard_ = RefPoint();
    emit clipboardAvailable( false );
}

//____________________________________________________________
void RefPointListWidget::add( const RefPoint& point )
{
    Debug::Throw( "RefPointListWidget::add.\n" );
    model_.add( point );
    list_->resizeColumns();
}

//____________________________________________________________
void RefPointListWidget::remove( const RefPoint& point )
{
    Debug::Throw( "RefPointListWidget::remove.\n" );
    model_.remove( point );
    list_->resizeColumns();
}

//____________________________________________________________
void RefPointListWidget::copy( const RefPoint& point )
{
    Debug::Throw( "RefPointListWidget::copy.\n" );
    clipboard_ = point;
    emit clipboardAvailable( true );
}

//____________________________________________________________
void RefPointListWidget::clearSelection( void )
{
    Debug::Throw( "RefPointListWidget::clearSelection.\n" );
    list_->clearSelection();
}

//____________________________________________________________
bool RefPointListWidget::select( const RefPoint& point )
{
    Debug::Throw( "RefPointListWidget::select.\n" );

    const QModelIndex index( model_.index( point ) );
    if( index.isValid() )
    {

        list_->clearSelection();
        list_->selectionModel()->select( index, QItemSelectionModel::Select|QItemSelectionModel::Current|QItemSelectionModel::Rows );
        return true;

    } else {

        // try find inactive point
        for( int index = 0; index < (int) dataSets_.size(); index++ )
        {

            // skip current data set
            if( index == dataSetIndex_ ) continue;
            const RefPoint::List& points( dataSets_[index] );
            RefPoint::List::const_iterator iter( std::find( points.begin(), points.end(), point ) );
            if( iter != points.end() )
            {
                setDataSetIndex( index );
                select( point );
                return true;
            }
        }

    }

    return false;

}

//____________________________________________________________
bool RefPointListWidget::replace( const RefPoint& source, const RefPoint& destination )
{

    Debug::Throw( "RefPointListWidget::replace.\n" );
    QModelIndex index( model_.index( source ) );
    if( index.isValid() )
    {
        model_.replace( index, destination );
        list_->resizeColumns();
        return true;
    } else return false;
}

//____________________________________________________________
void RefPointListWidget::updateActions( void )
{

    Debug::Throw( "RefPointListWidget::updateActions.\n" );

    QModelIndexList selection( list_->selectionModel()->selectedRows() );
    const bool hasSelection( !selection.empty() );

    const bool hasBefore( hasSelection && selection.front().row() > 0 );
    const bool hasAfter( hasSelection && selection.back().row() < model_.rowCount()-1 );
    removeAction_->setEnabled( hasSelection );
    moveUpAction_->setEnabled( hasBefore );
    moveDownAction_->setEnabled( hasAfter );

    // print action
    bool modelEmpty( model_.rowCount() == 0 );
    printAction_->setEnabled( !modelEmpty );
    printAction_->setText( dataSets_.size() > 1 ? tr( "Print Active Marker Set" ):tr( "Print Markers" ) );

    // print all action
    bool collectionEmpty( true );
    collectionEmpty &= modelEmpty;
    if( collectionEmpty )
    {
        for( int index = 0; index < (int) dataSets_.size(); index++ )
        {
            if( index == dataSetIndex_ ) continue;
            if( !dataSets_[index].empty() ) { collectionEmpty = false; break; }
        }
    }

    printAllAction_->setEnabled( !collectionEmpty );
    printAllAction_->setVisible( dataSets_.size() > 1 );

}

//____________________________________________________________
void RefPointListWidget::_updateContextMenu( const QPoint& position )
{

    Debug::Throw( "RefPointListWidget::updateErrorsActions.\n" );

    QModelIndex current( list_->selectionModel()->currentIndex() );
    const bool hasCurrent( current.isValid() );
    editErrorsAction_->setEnabled( hasCurrent );

    RefPoint currentPoint( current.isValid() ? model_.get( current ):RefPoint() );

    defaultErrorsAction_->setEnabled(
        hasCurrent &&
        ((!currentPoint.hasErrorX() && XmlOptions::get().get<double>( "DEFAULT_ERROR_X" ) > 0 ) ||
        (!currentPoint.hasErrorY() && XmlOptions::get().get<double>( "DEFAULT_ERROR_Y" ) > 0 ) )
        );

    defaultErrorsSystAction_->setEnabled(
        hasCurrent &&
        ((!currentPoint.hasErrorSystX() && XmlOptions::get().get<double>( "DEFAULT_ERROR_SYST_X" ) > 0 ) ||
        (!currentPoint.hasErrorSystY() && XmlOptions::get().get<double>( "DEFAULT_ERROR_SYST_Y" ) > 0 ) )
        );

    BaseContextMenu menu( list_ );
    menu.setHideDisabledActions( true );
    menu.addAction( removeAction_ );
    menu.addAction( moveUpAction_ );
    menu.addAction( moveDownAction_ );
    menu.addAction( sortAction_ );
    menu.addSeparator();
    menu.addAction( editErrorsAction_ );
    menu.addAction( defaultErrorsAction_ );
    menu.addAction( defaultErrorsSystAction_ );

    // execute
    menu.exec( list_->viewport()->mapToGlobal( position ) );

}

//____________________________________________________________
void RefPointListWidget::_moveUp( void )
{

    Debug::Throw( "RefPointListWidget::_moveUp.\n" );

    // retrieve selected indexes in list and store in model
    QModelIndexList selectedIndexes( list_->selectionModel()->selectedRows() );
    if( selectedIndexes.empty() ) return;

    RefPoint::List selectedRefPoints( model_.get( selectedIndexes ) );
    RefPoint::List currentRefPoints( model_.get() );
    RefPoint::List newRefPoints;

    for( RefPoint::List::const_iterator iter = currentRefPoints.begin(); iter != currentRefPoints.end(); ++iter )
    {

        // check if new list is not empty, current index is selected and last index is not.
        // if yes, move.
        if(
            !( newRefPoints.empty() ||
            selectedIndexes.indexOf( model_.index( *iter ) ) == -1 ||
            selectedIndexes.indexOf( model_.index( newRefPoints.back() ) ) != -1
            ) )
        {
            RefPoint last( newRefPoints.back() );
            newRefPoints.removeLast();
            newRefPoints << *iter;
            newRefPoints << last;
        } else newRefPoints << *iter;

    }

    model_.set( newRefPoints );

    // restore selection
    list_->selectionModel()->select( model_.index( selectedRefPoints.front() ),  QItemSelectionModel::Clear|QItemSelectionModel::Select|QItemSelectionModel::Rows );
    for( RefPoint::List::const_iterator iter = selectedRefPoints.begin(); iter != selectedRefPoints.end(); ++iter )
    { list_->selectionModel()->select( model_.index( *iter ), QItemSelectionModel::Select|QItemSelectionModel::Rows ); }

    return;

}

//____________________________________________________________
void RefPointListWidget::_moveDown( void )
{

    Debug::Throw( "RefPointListWidget::_moveDown.\n" );

    // retrieve selected indexes in list and store in model
    QModelIndexList selectedIndexes( list_->selectionModel()->selectedIndexes() );
    if( selectedIndexes.empty() ) return;

    RefPoint::List selectedRefPoints( model_.get( selectedIndexes ) );
    RefPoint::List newRefPoints;

    RefPoint::ListIterator iter( model_.get() );
    iter.toBack();
    while( iter.hasPrevious() )
    {
        const RefPoint& current( iter.previous() );
        if(
            !( newRefPoints.empty() ||
            selectedIndexes.indexOf( model_.index( current ) ) == -1 ||
            selectedIndexes.indexOf( model_.index( newRefPoints.front() ) ) != -1
            ) )
        {

            RefPoint first( newRefPoints.front() );
            newRefPoints.removeFirst();
            newRefPoints.push_front( current );
            newRefPoints.push_front( first );

        } else newRefPoints.push_front( current );
    }

    model_.set( newRefPoints );

    // restore selection
    list_->selectionModel()->select( model_.index( selectedRefPoints.front() ),  QItemSelectionModel::Clear|QItemSelectionModel::Select|QItemSelectionModel::Rows );
    for( RefPoint::List::const_iterator iter = selectedRefPoints.begin(); iter != selectedRefPoints.end(); ++iter )
    { list_->selectionModel()->select( model_.index( *iter ), QItemSelectionModel::Select|QItemSelectionModel::Rows ); }

}

//____________________________________________________________
void RefPointListWidget::_sort( void )
{
    Debug::Throw( "RefPointListWidget::_sort.\n" );

    // retrieve selected indexes in list and store in model
    QModelIndexList selectedIndexes( list_->selectionModel()->selectedIndexes() );
    RefPoint::List selectedRefPoints( model_.get( selectedIndexes ) );

    // store current ref points, sort and reset.
    RefPoint::List currentRefPoints( model_.get() );
    std::sort( currentRefPoints.begin(), currentRefPoints.end() );
    model_.set( currentRefPoints );

    // restore selection
    if( !selectedRefPoints.empty() )
    {
        list_->selectionModel()->select( model_.index( selectedRefPoints.front() ),  QItemSelectionModel::Clear|QItemSelectionModel::Select|QItemSelectionModel::Rows );
        for( RefPoint::List::const_iterator iter = selectedRefPoints.begin(); iter != selectedRefPoints.end(); ++iter )
        { list_->selectionModel()->select( model_.index( *iter ), QItemSelectionModel::Select|QItemSelectionModel::Rows ); }
    }

}

//____________________________________________________________
void RefPointListWidget::_remove( void )
{ emit selectionRemoved( model_.get( list_->selectionModel()->selectedRows() ) ); }

//____________________________________________________________
void RefPointListWidget::_print( void )
{

    Debug::Throw( "RefPointListWidget::_print.\n" );
    PrintListDialog dialog( this );
    dialog.setCollection( collection(), dataSetIndex_ );

    // connections
    connect( &dialog.formatToolButton(), SIGNAL( clicked( void ) ), SLOT( _configureRefPointFormat( void ) ) );
    connect( this, SIGNAL( refPointFormatChanged( void ) ), &dialog, SLOT( updateText( void ) ) );

    dialog.exec();

}

//____________________________________________________________
void RefPointListWidget::_printAll( void )
{

    Debug::Throw( "RefPointListWidget::_printAll.\n" );
    PrintListDialog dialog( this );
    dialog.setCollection( collection() );
    dialog.exec();

}

//____________________________________________________________
void RefPointListWidget::_editErrors( void )
{

    Debug::Throw( "RefPointListWidget::_editErrors.\n" );
    const QModelIndex index( list_->selectionModel()->currentIndex() );
    if( !index.isValid() ) return;
    emit errorsEdited( model_.get( index ) );

}

//____________________________________________________________
void RefPointListWidget::_addDefaultErrors( void )
{

    Debug::Throw( "RefPointListWidget::_addDefaultErrors.\n" );
    const QModelIndex index( list_->selectionModel()->currentIndex() );
    if( !index.isValid() ) return;
    emit defaultErrorsAdded( model_.get( index ) );

}

//____________________________________________________________
void RefPointListWidget::_addDefaultErrorsSyst( void )
{

    Debug::Throw( "RefPointListWidget::_addDefaultErrorsSyst.\n" );
    const QModelIndex index( list_->selectionModel()->currentIndex() );
    if( !index.isValid() ) return;
    emit defaultErrorsSystAdded( model_.get( index ) );

}

//____________________________________________________________
void RefPointListWidget::_addDataSet( void )
{

    Debug::Throw( "RefPointListWidget::_addDataSet.\n" );
    dataSets_ << RefPoint::List();
    dataSetWidget_->setCollection( dataSets_ );
    dataSetWidget_->select( dataSets_.size()-1 );
    updateActions();
    emit modified();

}

//____________________________________________________________
void RefPointListWidget::_removeDataSet( void )
{

    Debug::Throw( "RefPointListWidget::_removeDataSet.\n" );
    if( dataSetIndex_ >= (int) dataSets_.size() ) return;
    if( dataSets_.size() < 2 ) return;

    if( !QuestionDialog( this,
        tr( "Remove current data set ?\n"
        "Warning: all associated markers will be definitly lost" ) ).exec() ) return;

    RefPoint::Collection local( dataSets_ );
    dataSets_.clear();
    for( int index = 0; index < (int) local.size(); index++ )
    { if( index != dataSetIndex_ ) dataSets_ << local[index]; }

    const int oldIndex( dataSetIndex_ );

    dataSetWidget_->setCollection( dataSets_ );

    // update selection
    const int index( qMin( oldIndex, (int)dataSets_.size()-1 ) );
    dataSetWidget_->select( index );
    _selectDataSet( index );

    updateActions();
    emit modified();

}

//____________________________________________________________
void RefPointListWidget::_selectDataSet( int index )
{

    Debug::Throw() << "RefPointListWidget::_selectDataSet - index: " << index << endl;
    if( index < 0 || index >= (int) dataSets_.size() ) return;
    setDataSetIndex( index );
    emit modified();

}

//____________________________________________________________
void RefPointListWidget::_configureRefPointFormat( void )
{

    Debug::Throw( "RefPointListWidget::_configureRefPointFormat.\n" );
    RefPointFormatDialog dialog( this );

    // connections
    connect( &dialog.applyButton(), SIGNAL( clicked( void ) ), this, SIGNAL( refPointFormatChanged( void ) ) );
    connect( &dialog.okButton(), SIGNAL( clicked( void ) ), this, SIGNAL( refPointFormatChanged( void ) ) );

    dialog.centerOnWidget( qApp->activeWindow() ).exec();

}

//____________________________________________________________
void RefPointListWidget::_installActions( void )
{

    addAction( moveUpAction_ = new QAction( IconEngine::get( ICONS::UP ), tr( "Move Selection Up" ), this ) );
    connect( moveUpAction_, SIGNAL( triggered( void ) ), SLOT( _moveUp( void ) ) );

    addAction( moveDownAction_ = new QAction( IconEngine::get( ICONS::DOWN ), tr( "Move Selection Down" ), this ) );
    connect( moveDownAction_, SIGNAL( triggered( void ) ), SLOT( _moveDown( void ) ) );

    addAction( sortAction_ = new QAction( IconEngine::get( ICONS::SORT ), tr( "Sort Markers" ), this ) );
    connect( sortAction_, SIGNAL( triggered( void ) ), SLOT( _sort( void ) ) );

    addAction( removeAction_ = new QAction( IconEngine::get( ICONS::REMOVE ), tr( "Remove Selection" ), this ) );
    connect( removeAction_, SIGNAL( triggered( void ) ), SLOT( _remove( void ) ) );

    addAction( printAction_ = new QAction( IconEngine::get( ICONS::PRINT ),tr(  "Print Markers" ), this ) );
    connect( printAction_, SIGNAL( triggered( void ) ), SLOT( _print( void ) ) );

    addAction( printAllAction_ = new QAction( IconEngine::get( ICONS::PRINT ), tr( "Print All Markers" ), this ) );
    connect( printAllAction_, SIGNAL( triggered( void ) ), SLOT( _printAll( void ) ) );

    addAction( editErrorsAction_ = new QAction( IconEngine::get( ICONS::EDIT ), tr( "Edit Marker Errors" ), this ) );
    connect( editErrorsAction_, SIGNAL( triggered( void ) ), SLOT( _editErrors( void ) ) );

    addAction( defaultErrorsAction_ = new QAction( IconEngine::get( ICONS::EDIT ), tr( "Add Default Statistical Errors to Marker" ), this ) );
    connect( defaultErrorsAction_, SIGNAL( triggered( void ) ), SLOT( _addDefaultErrors( void ) ) );

    addAction( defaultErrorsSystAction_ = new QAction( IconEngine::get( ICONS::EDIT ), tr( "Add Default Systematic Errors to Marker" ), this ) );
    connect( defaultErrorsSystAction_, SIGNAL( triggered( void ) ), SLOT( _addDefaultErrorsSyst( void ) ) );

    addAction( refPointFormatAction_ = new QAction( IconEngine::get( ICONS::CONFIGURE ), tr( "Configure Marker Format for Printing..." ), this ) );
    connect( refPointFormatAction_, SIGNAL( triggered( void ) ), SLOT( _configureRefPointFormat( void ) ) );

}

//______________________________________
void RefPointListWidget::_updateConfiguration( void )
{
    Debug::Throw( "RefPointListWidget::_updateConfiguration.\n" );
    model_.setGlobalFormat( XmlOptions::get().get<QString>( "GLOBAL_FORMAT" ) );
    model_.setLocalFormat( XmlOptions::get().get<QString>( "LOCAL_FORMAT" ) );
    model_.setErrorFormat( XmlOptions::get().get<QString>( "ERROR_FORMAT" ) );
    list_->resizeColumns();
}
