// $Id: RefPointModel.cpp,v 1.14 2013/03/22 09:41:40 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "RefPointModel.h"
#include "Debug.h"

#include <QPalette>
#include <QTextStream>

//_______________________________________________
const QString RefPointModel::columnTitles_[ RefPointModel::nColumns ] =
{
    tr( "Index" ),
    tr( "Global" ),
    tr( "Local" ),
    tr( "Stat. Errors (x)" ),
    tr( "Stat. Errors (y)" ),
    tr( "Syst. Errors (x)" ),
    tr( "Syst. Errors (y)" )
};

//__________________________________________________________________
QVariant RefPointModel::data( const QModelIndex& index, int role ) const
{

    // check index, role and column
    if( !index.isValid() ) return QVariant();

    // retrieve associated file info
    const RefPoint& point( get(index) );

    // return text associated to file and column
    if( role == Qt::DisplayRole ) {

        switch( index.column() )
        {

            case INDEX: return index.row();

            case GLOBAL: return point.global().toString( globalFormat_ );

            case LOCAL: return point.local().toString( localFormat_ );

            case ERRORS_X:
            if( point.hasErrorX() ) return error( point.errorX(), point.local().x() );
            else return QVariant();

            case ERRORS_Y:
            if( point.hasErrorY() ) return error( point.errorY(), point.local().y() );
            else return QVariant();

            case ERRORS_SYST_X:
            if( point.hasErrorSystX() ) return error( point.errorSystX(), point.local().x() );
            else return QVariant();

            case ERRORS_SYST_Y:
            if( point.hasErrorSystY() ) return error( point.errorSystY(), point.local().y() );
            else return QVariant();

            default:
            return QVariant();

        }

    } else if( role == Qt::ForegroundRole ) {

        if( !refPointsValid_ ) return QPalette().color( QPalette::Disabled, QPalette::Text );
        else return QVariant();

    }

    return QVariant();

}

//__________________________________________________________________
QVariant RefPointModel::headerData(int section, Qt::Orientation orientation, int role) const
{

    if(
        orientation == Qt::Horizontal &&
        role == Qt::DisplayRole &&
        section >= 0 &&
        section < nColumns )
    { return columnTitles_[section]; }

    // return empty
    return QVariant();

}

//__________________________________________________________________
QString RefPointModel::error( const RefPoint::Error& error, const double& value ) const
{

    double plus( error.plus_ );
    double minus( error.minus_ );
    if( error.relative_ )
    {
        plus *= value;
        minus *= value;
    }

    if( errorFormat_.isEmpty() )
    {

        return QString( "( +%1, -%2 )" ).arg( plus ).arg( minus );

    } else {

        return QString( "( +%1, -%2 )" )
            .arg( QString().sprintf( errorFormat_.toLatin1().constData(), plus ) )
            .arg( QString().sprintf( errorFormat_.toLatin1().constData(), minus ) );

    }

}

//__________________________________________________________________
void RefPointModel::setGlobalFormat( const QString& format )
{
    // store format
    globalFormat_ = format;
    emit layoutChanged();

}

//__________________________________________________________________
void RefPointModel::setLocalFormat( const QString& format )
{
    // store format
    localFormat_ = format;
    emit layoutChanged();

}

//__________________________________________________________________
void RefPointModel::setErrorFormat( const QString& format )
{
    // store format
    errorFormat_ = format;
    emit layoutChanged();

}
