#ifndef RefPointPanel_h
#define RefPointPanel_h

// $Id: RefPointPanel.h,v 1.33 2013/03/22 09:41:40 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "Counter.h"
#include "RefPointWidget.h"

#include <QAbstractItemView>
#include <QActionGroup>
#include <QCheckBox>
#include <QFrame>
#include <QLayout>
#include <QPainter>
#include <QPaintEvent>
#include <QWidget>
#include <QVector>

class ZoomWidget;

class RefPointPanel: public QWidget, public Counter
{

    Q_OBJECT

    public:

    //! number of ref points
    enum { RefPointCount = 3 };

    //! constructor
    RefPointPanel( QWidget* = 0 );

    //! destructor
    virtual ~RefPointPanel( void )
    {}

    //!@name accessors
    //@{

    //! button group
    QActionGroup& editActions( void ) const
    { return *editActions_; }

    //! button group
    QActionGroup& clearActions( void ) const
    { return *clearActions_; }

    // has hover
    bool hasHover( void ) const
    {
        for( int index=0; index<RefPointCount; index++ )
        { if( refPointWidgets_[index]->isHovered() ) return true; }

        return false;
    }

    // has selection
    bool hasSelection( void ) const
    {
        for( int index=0; index<RefPointCount; index++ )
        { if( refPointWidgets_[index]->isSelected() ) return true; }

        return false;
    }

    //! selected point
    RefPoint selection( void ) const
    {
        for( int index=0; index<RefPointCount; index++ )
        { if( refPointWidgets_[index]->isSelected() ) return refPointWidgets_[index]->point(); }
        return RefPoint();
    }

    //! selected index
    int selectedIndex( void ) const
    {
        for( int index=0; index<RefPointCount; index++ )
        { if( refPointWidgets_[index]->isSelected() ) return index; }
        return -1;
    }

    //! get point for index
    const RefPoint& point( int index ) const
    { return refPointWidgets_[index]->point(); }

    //! get index for point
    int index( const RefPoint& point ) const
    {

        for( int index = 0; index < RefPointCount; ++index )
        { if( refPointWidgets_[index]->point() == point ) return index; }

        return -1;

    }

    //! get list of reference points
    RefPoint::List points( void ) const
    {
        RefPoint::List out;
        for( int index=0; index<RefPointCount; index++ )
        { if( refPointWidgets_[index]->isValid() ) out << point(index); }

        return out;
    }

    //! true when valid
    bool isValid( void ) const
    { return validate() == Valid; }

    //! get first invalid index
    int firstInvalidIndex( void ) const;

    //! true if refPoint is in list
    bool contains( const RefPoint& point ) const
    { return index( point ) >= 0; }

    //! transformation
    RefPoint::PointF toLocal( const RefPoint::PointF& ) const;

    //! transformation
    RefPoint::PointF toGlobal( const RefPoint::PointF& ) const;

    //! transformation
    RefPoint toLocal( const RefPoint& ) const;

    //! transformation
    RefPoint toGlobal( const RefPoint& ) const;

    //! true if log x selected
    bool hasLogX( void ) const
    { return logXCheckBox_->isChecked(); }

    //! true if log x selected
    bool hasLogY( void ) const
    { return logYCheckBox_->isChecked(); }

    //! zoom widget
    ZoomWidget& zoomWidget( void ) const
    { return *zoomWidget_; }

    //@}

    //!@name modifiers
    //@{

    //! select
    void select( const RefPoint& );

    //! select index
    bool select( int );

    //! clear selection
    void clearSelection( void );

    //! set reference point at index
    void setPoint( int index, const RefPoint& refPoint );

    //! validation error code
    enum ErrorCode
    {
        Valid,
        Incomplete,
        LogXInvalid,
        LogYInvalid,
        GlobalInvalid,
        LocalInvalid
    };

    ErrorCode validate( void ) const;

    //! toggle log x
    void setLogX( bool value )
    { logXCheckBox_->setChecked( value ); }

    //! toggle log y
    void setLogY( bool value )
    { logYCheckBox_->setChecked( value ); }

    //! clear
    void clear( int index )
    { refPointWidgets_[index]->clear(); }

    //! clear
    void clear( void )
    {
        clearSelection();
        foreach( RefPointWidget* widget, refPointWidgets_ )
        { widget->clear(); }
    }

    //! remove
    bool remove( const RefPoint& );

    //! event filter (used to select refpoint by left clicking
    bool eventFilter( QObject*, QEvent* );

    //@}

    signals:

    //! emmited when changed
    void modified( void );

    //! emmited when selection has changed
    void selectionChanged( void );

    protected slots:

    //! update transormation
    void _updateTransformation( void );

    protected:

    //! paint event
    virtual void paintEvent( QPaintEvent* );

    //! mouse press event
    virtual void mousePressEvent( QMouseEvent* event )
    {
        clearSelection();
        emit selectionChanged();
        return QWidget::mousePressEvent( event );
    }

    //! determinant (global coordinates)
    double _globalDeterminant( void ) const;

    //! determinant (local coordinates)
    double _localDeterminant( void ) const;

    private:

    //! some styles require an item view passed to painting method to have proper selection rendered in items
    QAbstractItemView* itemView_;

    //! button group
    QActionGroup* editActions_;

    //! button group
    QActionGroup* clearActions_;

    //! zoom Widget
    ZoomWidget* zoomWidget_;

    //!@name checkboxes
    //@{
    QCheckBox* logXCheckBox_;
    QCheckBox* logYCheckBox_;
    //@}

    //! list of refpoints
    RefPointWidget::List refPointWidgets_;

    //!@name transformation elements
    //@{

    //! origin
    QPointF origin_;

    //! matrix (a00, a01, a10, a11 )
    QVector<double> matrix_;

    //@}

};

#endif
